MODULE icedyn_rhg_eap
   !!======================================================================
   !!                     ***  MODULE  icedyn_rhg_eap  ***
   !!   Sea-Ice dynamics : rheology Elasto-Viscous-Plastic
   !!======================================================================
   !! History :   -   !  2007-03  (M.A. Morales Maqueda, S. Bouillon) Original code
   !!            3.0  !  2008-03  (M. Vancoppenolle) adaptation to new model
   !!             -   !  2008-11  (M. Vancoppenolle, S. Bouillon, Y. Aksenov) add surface tilt in ice rheolohy 
   !!            3.3  !  2009-05  (G.Garric)    addition of the evp case
   !!            3.4  !  2011-01  (A. Porter)   dynamical allocation 
   !!            3.5  !  2012-08  (R. Benshila) AGRIF
   !!            3.6  !  2016-06  (C. Rousset)  Rewriting + landfast ice + mEVP (Bouillon 2013)
   !!            3.7  !  2017     (C. Rousset)  add aEVP (Kimmritz 2016-2017)
   !!            4.0  !  2018     (many people) SI3 [aka Sea Ice cube]
   !!                 !  2019     (S. Rynders)  change into eap rheology from
   !!                                           CICE code (Tsamados, Heorton) 
   !!----------------------------------------------------------------------
#if defined key_si3
   !!----------------------------------------------------------------------
   !!   'key_si3'                                       SI3 sea-ice model
   !!----------------------------------------------------------------------
   !!   ice_dyn_rhg_eap : computes ice velocities from EVP rheology
   !!   rhg_eap_rst     : read/write EVP fields in ice restart
   !!----------------------------------------------------------------------
   USE phycst         ! Physical constant
   USE dom_oce        ! Ocean domain
   USE sbc_oce , ONLY : ln_ice_embd, nn_fsbc, ssh_m
   USE sbc_ice , ONLY : utau_ice, vtau_ice, snwice_mass, snwice_mass_b
   USE ice            ! sea-ice: ice variables
   USE icevar         ! ice_var_sshdyn
   USE icedyn_rdgrft  ! sea-ice: ice strength
   USE bdy_oce , ONLY : ln_bdy 
   USE bdyice 
#if defined key_agrif
   USE agrif_ice_interp
#endif
   !
   USE in_out_manager ! I/O manager
   USE iom            ! I/O manager library
   USE lib_mpp        ! MPP library
   USE lib_fortran    ! fortran utilities (glob_sum + no signed zero)
   USE lbclnk         ! lateral boundary conditions (or mpp links)
   USE prtctl         ! Print control

   IMPLICIT NONE
   PRIVATE

   PUBLIC   ice_dyn_rhg_eap   ! called by icedyn_rhg.F90
   PUBLIC   rhg_eap_rst       ! called by icedyn_rhg.F90

   REAL(wp), PARAMETER           ::   phi = 3.141592653589793_wp/12._wp    ! diamond shaped floe smaller angle (default phi = 30 deg)

   ! look-up table for calculating structure tensor
   INTEGER, PARAMETER ::   nx_yield = 41
   INTEGER, PARAMETER ::   ny_yield = 41
   INTEGER, PARAMETER ::   na_yield = 21

   REAL(wp), DIMENSION(nx_yield, ny_yield, na_yield) ::   s11r, s12r, s22r, s11s, s12s, s22s

   !! * Substitutions
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !! NEMO/ICE 4.0 , NEMO Consortium (2018)
   !! $Id: icedyn_rhg_eap.F90 11536 2019-09-11 13:54:18Z smasson $
   !! Software governed by the CeCILL license (see ./LICENSE)
   !!----------------------------------------------------------------------
CONTAINS

   SUBROUTINE ice_dyn_rhg_eap( kt, pstress1_i, pstress2_i, pstress12_i, pshear_i, pdivu_i, pdelta_i, aniso_11, aniso_12 )
      !!-------------------------------------------------------------------
      !!                 ***  SUBROUTINE ice_dyn_rhg_eap  ***
      !!                             EAP-C-grid
      !!
      !! ** purpose : determines sea ice drift from wind stress, ice-ocean
      !!  stress and sea-surface slope. Ice-ice interaction is described by 
      !!  a non-linear elasto-anisotropic-plastic (EAP) law including shear 
      !!  strength and a bulk rheology .	
      !!
      !!  The points in the C-grid look like this, dear reader
      !!
      !!                              (ji,jj)
      !!                                 |
      !!                                 |
      !!                      (ji-1,jj)  |  (ji,jj)
      !!                             ---------   
      !!                            |         |
      !!                            | (ji,jj) |------(ji,jj)
      !!                            |         |
      !!                             ---------   
      !!                     (ji-1,jj-1)     (ji,jj-1)
      !!
      !! ** Inputs  : - wind forcing (stress), oceanic currents
      !!                ice total volume (vt_i) per unit area
      !!                snow total volume (vt_s) per unit area
      !!
      !! ** Action  : - compute u_ice, v_ice : the components of the 
      !!                sea-ice velocity vector
      !!              - compute delta_i, shear_i, divu_i, which are inputs
      !!                of the ice thickness distribution
      !!
      !! ** Steps   : 0) compute mask at F point
      !!              1) Compute ice snow mass, ice strength 
      !!              2) Compute wind, oceanic stresses, mass terms and
      !!                 coriolis terms of the momentum equation
      !!              3) Solve the momentum equation (iterative procedure)
      !!              4) Recompute delta, shear and divergence
      !!                 (which are inputs of the ITD) & store stress
      !!                 for the next time step
      !!              5) Diagnostics including charge ellipse
      !!
      !! ** Notes   : There is the possibility to use aEVP from the nice work of Kimmritz et al. (2016 & 2017)
      !!              by setting up ln_aEVP=T (i.e. changing alpha and beta parameters).
      !!              This is an upgraded version of mEVP from Bouillon et al. 2013
      !!              (i.e. more stable and better convergence)
      !!
      !! References : Hunke and Dukowicz, JPO97
      !!              Bouillon et al., Ocean Modelling 2009
      !!              Bouillon et al., Ocean Modelling 2013
      !!              Kimmritz et al., Ocean Modelling 2016 & 2017
      !!-------------------------------------------------------------------
      INTEGER                 , INTENT(in   ) ::   kt                                    ! time step
      REAL(wp), DIMENSION(:,:), INTENT(inout) ::   pstress1_i, pstress2_i, pstress12_i   !
      REAL(wp), DIMENSION(:,:), INTENT(  out) ::   pshear_i  , pdivu_i   , pdelta_i      !
      REAL(wp), DIMENSION(:,:), INTENT(inout) ::   aniso_11, aniso_12   ! structure tensor components
      !!
      INTEGER ::   ji, jj       ! dummy loop indices
      INTEGER ::   jter         ! local integers
      !
      REAL(wp) ::   zrhoco                                              ! rau0 * rn_cio
      REAL(wp) ::   zdtevp, z1_dtevp                                    ! time step for subcycling
      REAL(wp) ::   ecc2, z1_ecc2                                       ! square of yield ellipse eccenticity
      REAL(wp) ::   zalph1, z1_alph1                                    ! alpha coef from Bouillon 2009 or Kimmritz 2017
      REAL(wp) ::   zm1, zm2, zm3, zmassU, zmassV, zvU, zvV             ! ice/snow mass and volume
      REAL(wp) ::   zdelta, zp_delf, zds2, zdt, zdt2, zdiv, zdiv2, zdsT ! temporary scalars
      REAL(wp) ::   zTauO, zTauB, zRHS, zvel                            ! temporary scalars
      REAL(wp) ::   zkt                                                 ! isotropic tensile strength for landfast ice
      REAL(wp) ::   zvCr                                                ! critical ice volume above which ice is landfast
      !
      REAL(wp) ::   zresm                                               ! Maximal error on ice velocity
      REAL(wp) ::   zintb, zintn                                        ! dummy argument
      REAL(wp) ::   zfac_x, zfac_y
      REAL(wp) ::   zshear, zdum1, zdum2
      REAL(wp) ::   stressptmp, stressmtmp                              ! anisotropic stress tensor components
      REAL(wp) ::   alphar, alphas                                      ! for mechanical redistribution
      !
      REAL(wp), DIMENSION(jpi,jpj) ::   stress12tmp                     ! anisotropic stress tensor component for regridding
      REAL(wp), DIMENSION(jpi,jpj) ::   zp_delt                         ! P/delta at T points
      REAL(wp), DIMENSION(jpi,jpj) ::   zbeta                           ! beta coef from Kimmritz 2017
      !
      REAL(wp), DIMENSION(jpi,jpj) ::   zdt_m                           ! (dt / ice-snow_mass) on T points
      REAL(wp), DIMENSION(jpi,jpj) ::   zaU  , zaV                      ! ice fraction on U/V points
      REAL(wp), DIMENSION(jpi,jpj) ::   zmU_t, zmV_t                    ! (ice-snow_mass / dt) on U/V points
      REAL(wp), DIMENSION(jpi,jpj) ::   zmf                             ! coriolis parameter at T points
      REAL(wp), DIMENSION(jpi,jpj) ::   v_oceU, u_oceV, v_iceU, u_iceV  ! ocean/ice u/v component on V/U points                           
      !
      REAL(wp), DIMENSION(jpi,jpj) ::   zds                             ! shear
      REAL(wp), DIMENSION(jpi,jpj) ::   zs1, zs2, zs12                  ! stress tensor components
!!$      REAL(wp), DIMENSION(jpi,jpj) ::   zu_ice, zv_ice, zresr           ! check convergence
      REAL(wp), DIMENSION(jpi,jpj) ::   zsshdyn                         ! array used for the calculation of ice surface slope:
      !                                                                 !    ocean surface (ssh_m) if ice is not embedded
      !                                                                 !    ice bottom surface if ice is embedded   
      REAL(wp), DIMENSION(jpi,jpj) ::   zfU  , zfV                      ! internal stresses
      REAL(wp), DIMENSION(jpi,jpj) ::   zspgU, zspgV                    ! surface pressure gradient at U/V points
      REAL(wp), DIMENSION(jpi,jpj) ::   zCorU, zCorV                    ! Coriolis stress array
      REAL(wp), DIMENSION(jpi,jpj) ::   ztaux_ai, ztauy_ai              ! ice-atm. stress at U-V points
      REAL(wp), DIMENSION(jpi,jpj) ::   ztaux_oi, ztauy_oi              ! ice-ocean stress at U-V points
      REAL(wp), DIMENSION(jpi,jpj) ::   ztaux_bi, ztauy_bi              ! ice-OceanBottom stress at U-V points (landfast)
      REAL(wp), DIMENSION(jpi,jpj) ::   ztaux_base, ztauy_base          ! ice-bottom stress at U-V points (landfast)
      !
      REAL(wp), DIMENSION(jpi,jpj) ::   zmsk01x, zmsk01y                ! dummy arrays
      REAL(wp), DIMENSION(jpi,jpj) ::   zmsk00x, zmsk00y                ! mask for ice presence
      REAL(wp), DIMENSION(jpi,jpj) ::   zfmask, zwf                     ! mask at F points for the ice

      REAL(wp), PARAMETER          ::   zepsi  = 1.0e-20_wp             ! tolerance parameter
      REAL(wp), PARAMETER          ::   zmmin  = 1._wp                  ! ice mass (kg/m2)  below which ice velocity becomes very small
      REAL(wp), PARAMETER          ::   zamin  = 0.001_wp               ! ice concentration below which ice velocity becomes very small
      !! --- diags
      REAL(wp), DIMENSION(jpi,jpj) ::   zmsk00
      REAL(wp), ALLOCATABLE, DIMENSION(:,:) ::   zsig1, zsig2, zsig3
      !! --- SIMIP diags
      REAL(wp), ALLOCATABLE, DIMENSION(:,:) ::   zdiag_xmtrp_ice ! X-component of ice mass transport (kg/s)
      REAL(wp), ALLOCATABLE, DIMENSION(:,:) ::   zdiag_ymtrp_ice ! Y-component of ice mass transport (kg/s)
      REAL(wp), ALLOCATABLE, DIMENSION(:,:) ::   zdiag_xmtrp_snw ! X-component of snow mass transport (kg/s)
      REAL(wp), ALLOCATABLE, DIMENSION(:,:) ::   zdiag_ymtrp_snw ! Y-component of snow mass transport (kg/s)
      REAL(wp), ALLOCATABLE, DIMENSION(:,:) ::   zdiag_xatrp     ! X-component of area transport (m2/s)
      REAL(wp), ALLOCATABLE, DIMENSION(:,:) ::   zdiag_yatrp     ! Y-component of area transport (m2/s)      
      !!-------------------------------------------------------------------

      IF( kt == nit000 .AND. lwp )   WRITE(numout,*) '-- ice_dyn_rhg_eap: EAP sea-ice rheology'
      !
!!gm for Clem:  OPTIMIZATION:  I think zfmask can be computed one for all at the initialization....
      !------------------------------------------------------------------------------!
      ! 0) mask at F points for the ice
      !------------------------------------------------------------------------------!
      ! ocean/land mask
      DO jj = 1, jpjm1
         DO ji = 1, jpim1      ! NO vector opt.
            zfmask(ji,jj) = tmask(ji,jj,1) * tmask(ji+1,jj,1) * tmask(ji,jj+1,1) * tmask(ji+1,jj+1,1)
         END DO
      END DO
      CALL lbc_lnk( 'icedyn_rhg_eap', zfmask, 'F', 1._wp )

      ! Lateral boundary conditions on velocity (modify zfmask)
      zwf(:,:) = zfmask(:,:)
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1   ! vector opt.
            IF( zfmask(ji,jj) == 0._wp ) THEN
               zfmask(ji,jj) = rn_ishlat * MIN( 1._wp , MAX( zwf(ji+1,jj), zwf(ji,jj+1), zwf(ji-1,jj), zwf(ji,jj-1) ) )
            ENDIF
         END DO
      END DO
      DO jj = 2, jpjm1
         IF( zfmask(1,jj) == 0._wp ) THEN
            zfmask(1  ,jj) = rn_ishlat * MIN( 1._wp , MAX( zwf(2,jj), zwf(1,jj+1), zwf(1,jj-1) ) )
         ENDIF
         IF( zfmask(jpi,jj) == 0._wp ) THEN
            zfmask(jpi,jj) = rn_ishlat * MIN( 1._wp , MAX( zwf(jpi,jj+1), zwf(jpim1,jj), zwf(jpi,jj-1) ) )
         ENDIF
      END DO
      DO ji = 2, jpim1
         IF( zfmask(ji,1) == 0._wp ) THEN
            zfmask(ji,1  ) = rn_ishlat * MIN( 1._wp , MAX( zwf(ji+1,1), zwf(ji,2), zwf(ji-1,1) ) )
         ENDIF
         IF( zfmask(ji,jpj) == 0._wp ) THEN
            zfmask(ji,jpj) = rn_ishlat * MIN( 1._wp , MAX( zwf(ji+1,jpj), zwf(ji-1,jpj), zwf(ji,jpjm1) ) )
         ENDIF
      END DO
      CALL lbc_lnk( 'icedyn_rhg_eap', zfmask, 'F', 1._wp )

      !------------------------------------------------------------------------------!
      ! 1) define some variables and initialize arrays
      !------------------------------------------------------------------------------!
      zrhoco = rau0 * rn_cio 

      ! ecc2: square of yield ellipse eccenticrity
      ecc2    = rn_ecc * rn_ecc
      z1_ecc2 = 1._wp / ecc2

      ! Time step for subcycling
      zdtevp   = rdt_ice / REAL( nn_nevp )
      z1_dtevp = 1._wp / zdtevp

      ! alpha parameters (Bouillon 2009)
      IF( .NOT. ln_aEVP ) THEN
         zalph1 = ( 2._wp * rn_relast * rdt_ice ) * z1_dtevp
         z1_alph1 = 1._wp / ( zalph1 + 1._wp )
      ENDIF
         
      ! Initialise stress tensor 
      zs1 (:,:) = pstress1_i (:,:) 
      zs2 (:,:) = pstress2_i (:,:)
      zs12(:,:) = pstress12_i(:,:)

      ! Ice strength
      CALL ice_strength

      ! landfast param from Lemieux(2016): add isotropic tensile strength (following Konig Beatty and Holland, 2010)
      IF( ln_landfast_L16 ) THEN   ;   zkt = rn_tensile
      ELSE                         ;   zkt = 0._wp
      ENDIF
      !
      !------------------------------------------------------------------------------!
      ! 2) Wind / ocean stress, mass terms, coriolis terms
      !------------------------------------------------------------------------------!
      ! sea surface height
      !    embedded sea ice: compute representative ice top surface
      !    non-embedded sea ice: use ocean surface for slope calculation
      zsshdyn(:,:) = ice_var_sshdyn( ssh_m, snwice_mass, snwice_mass_b)

      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1

            ! ice fraction at U-V points
            zaU(ji,jj) = 0.5_wp * ( at_i(ji,jj) * e1e2t(ji,jj) + at_i(ji+1,jj) * e1e2t(ji+1,jj) ) * r1_e1e2u(ji,jj) * umask(ji,jj,1)
            zaV(ji,jj) = 0.5_wp * ( at_i(ji,jj) * e1e2t(ji,jj) + at_i(ji,jj+1) * e1e2t(ji,jj+1) ) * r1_e1e2v(ji,jj) * vmask(ji,jj,1)

            ! Ice/snow mass at U-V points
            zm1 = ( rhos * vt_s(ji  ,jj  ) + rhoi * vt_i(ji  ,jj  ) )
            zm2 = ( rhos * vt_s(ji+1,jj  ) + rhoi * vt_i(ji+1,jj  ) )
            zm3 = ( rhos * vt_s(ji  ,jj+1) + rhoi * vt_i(ji  ,jj+1) )
            zmassU = 0.5_wp * ( zm1 * e1e2t(ji,jj) + zm2 * e1e2t(ji+1,jj) ) * r1_e1e2u(ji,jj) * umask(ji,jj,1)
            zmassV = 0.5_wp * ( zm1 * e1e2t(ji,jj) + zm3 * e1e2t(ji,jj+1) ) * r1_e1e2v(ji,jj) * vmask(ji,jj,1)

            ! Ocean currents at U-V points
            v_oceU(ji,jj)   = 0.25_wp * ( v_oce(ji,jj) + v_oce(ji,jj-1) + v_oce(ji+1,jj) + v_oce(ji+1,jj-1) ) * umask(ji,jj,1)
            u_oceV(ji,jj)   = 0.25_wp * ( u_oce(ji,jj) + u_oce(ji-1,jj) + u_oce(ji,jj+1) + u_oce(ji-1,jj+1) ) * vmask(ji,jj,1)

            ! Coriolis at T points (m*f)
            zmf(ji,jj)      = zm1 * ff_t(ji,jj)

            ! dt/m at T points (for alpha and beta coefficients)
            zdt_m(ji,jj)    = zdtevp / MAX( zm1, zmmin )
            
            ! m/dt
            zmU_t(ji,jj)    = zmassU * z1_dtevp
            zmV_t(ji,jj)    = zmassV * z1_dtevp
            
            ! Drag ice-atm.
            ztaux_ai(ji,jj) = zaU(ji,jj) * utau_ice(ji,jj)
            ztauy_ai(ji,jj) = zaV(ji,jj) * vtau_ice(ji,jj)

            ! Surface pressure gradient (- m*g*GRAD(ssh)) at U-V points
            zspgU(ji,jj)    = - zmassU * grav * ( zsshdyn(ji+1,jj) - zsshdyn(ji,jj) ) * r1_e1u(ji,jj)
            zspgV(ji,jj)    = - zmassV * grav * ( zsshdyn(ji,jj+1) - zsshdyn(ji,jj) ) * r1_e2v(ji,jj)

            ! masks
            zmsk00x(ji,jj) = 1._wp - MAX( 0._wp, SIGN( 1._wp, -zmassU ) )  ! 0 if no ice
            zmsk00y(ji,jj) = 1._wp - MAX( 0._wp, SIGN( 1._wp, -zmassV ) )  ! 0 if no ice

            ! switches
            IF( zmassU <= zmmin .AND. zaU(ji,jj) <= zamin ) THEN   ;   zmsk01x(ji,jj) = 0._wp
            ELSE                                                   ;   zmsk01x(ji,jj) = 1._wp   ;   ENDIF
            IF( zmassV <= zmmin .AND. zaV(ji,jj) <= zamin ) THEN   ;   zmsk01y(ji,jj) = 0._wp
            ELSE                                                   ;   zmsk01y(ji,jj) = 1._wp   ;   ENDIF

         END DO
      END DO
      CALL lbc_lnk_multi( 'icedyn_rhg_eap', zmf, 'T', 1., zdt_m, 'T', 1. )
      !
      !                                  !== Landfast ice parameterization ==!
      !
      IF( ln_landfast_L16 ) THEN         !-- Lemieux 2016
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               ! ice thickness at U-V points
               zvU = 0.5_wp * ( vt_i(ji,jj) * e1e2t(ji,jj) + vt_i(ji+1,jj) * e1e2t(ji+1,jj) ) * r1_e1e2u(ji,jj) * umask(ji,jj,1)
               zvV = 0.5_wp * ( vt_i(ji,jj) * e1e2t(ji,jj) + vt_i(ji,jj+1) * e1e2t(ji,jj+1) ) * r1_e1e2v(ji,jj) * vmask(ji,jj,1)
               ! ice-bottom stress at U points
               zvCr = zaU(ji,jj) * rn_depfra * hu_n(ji,jj)
               ztaux_base(ji,jj) = - rn_icebfr * MAX( 0._wp, zvU - zvCr ) * EXP( -rn_crhg * ( 1._wp - zaU(ji,jj) ) )
               ! ice-bottom stress at V points
               zvCr = zaV(ji,jj) * rn_depfra * hv_n(ji,jj)
               ztauy_base(ji,jj) = - rn_icebfr * MAX( 0._wp, zvV - zvCr ) * EXP( -rn_crhg * ( 1._wp - zaV(ji,jj) ) )
               ! ice_bottom stress at T points
               zvCr = at_i(ji,jj) * rn_depfra * ht_n(ji,jj)
               tau_icebfr(ji,jj) = - rn_icebfr * MAX( 0._wp, vt_i(ji,jj) - zvCr ) * EXP( -rn_crhg * ( 1._wp - at_i(ji,jj) ) )
            END DO
         END DO
         CALL lbc_lnk( 'icedyn_rhg_eap', tau_icebfr(:,:), 'T', 1. )
         !
      ELSE                               !-- no landfast
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               ztaux_base(ji,jj) = 0._wp
               ztauy_base(ji,jj) = 0._wp
            END DO
         END DO
      ENDIF

      !------------------------------------------------------------------------------!
      ! 3) Solution of the momentum equation, iterative procedure
      !------------------------------------------------------------------------------!
      !
      !                                               ! ==================== !
      DO jter = 1 , nn_nevp                           !    loop over jter    !
         !                                            ! ==================== !        
         l_full_nf_update = jter == nn_nevp   ! false: disable full North fold update (performances) for iter = 1 to nn_nevp-1
         !
!!$         IF(ln_ctl) THEN   ! Convergence test
!!$            DO jj = 1, jpjm1
!!$               zu_ice(:,jj) = u_ice(:,jj) ! velocity at previous time step
!!$               zv_ice(:,jj) = v_ice(:,jj)
!!$            END DO
!!$         ENDIF

         ! --- divergence, tension & shear (Appendix B of Hunke & Dukowicz, 2002) --- !
         DO jj = 1, jpjm1         ! loops start at 1 since there is no boundary condition (lbc_lnk) at i=1 and j=1 for F points
            DO ji = 1, jpim1

               ! shear at F points
               zds(ji,jj) = ( ( u_ice(ji,jj+1) * r1_e1u(ji,jj+1) - u_ice(ji,jj) * r1_e1u(ji,jj) ) * e1f(ji,jj) * e1f(ji,jj)   &
                  &         + ( v_ice(ji+1,jj) * r1_e2v(ji+1,jj) - v_ice(ji,jj) * r1_e2v(ji,jj) ) * e2f(ji,jj) * e2f(ji,jj)   &
                  &         ) * r1_e1e2f(ji,jj) * zfmask(ji,jj)

            END DO
         END DO
         CALL lbc_lnk( 'icedyn_rhg_eap', zds, 'F', 1. )

         DO jj = 2, jpj    ! loop to jpi,jpj to avoid making a communication for zs1,zs2,zs12
            DO ji = 2, jpi ! no vector loop

               ! shear at T points 
               zdsT = ( zds(ji,jj  ) * e1e2f(ji,jj  ) + zds(ji-1,jj  ) * e1e2f(ji-1,jj  )  &
                  &   + zds(ji,jj-1) * e1e2f(ji,jj-1) + zds(ji-1,jj-1) * e1e2f(ji-1,jj-1)  &
                  &   ) * 0.25_wp * r1_e1e2t(ji,jj)

               ! shear**2 at T points (doc eq. A16)
               zds2 = ( zds(ji,jj  ) * zds(ji,jj  ) * e1e2f(ji,jj  ) + zds(ji-1,jj  ) * zds(ji-1,jj  ) * e1e2f(ji-1,jj  )  &
                  &   + zds(ji,jj-1) * zds(ji,jj-1) * e1e2f(ji,jj-1) + zds(ji-1,jj-1) * zds(ji-1,jj-1) * e1e2f(ji-1,jj-1)  &
                  &   ) * 0.25_wp * r1_e1e2t(ji,jj)
              
               ! divergence at T points
               zdiv  = ( e2u(ji,jj) * u_ice(ji,jj) - e2u(ji-1,jj) * u_ice(ji-1,jj)   &
                  &    + e1v(ji,jj) * v_ice(ji,jj) - e1v(ji,jj-1) * v_ice(ji,jj-1)   &
                  &    ) * r1_e1e2t(ji,jj)
               zdiv2 = zdiv * zdiv
               
               ! tension at T points
               zdt  = ( ( u_ice(ji,jj) * r1_e2u(ji,jj) - u_ice(ji-1,jj) * r1_e2u(ji-1,jj) ) * e2t(ji,jj) * e2t(ji,jj)   &
                  &   - ( v_ice(ji,jj) * r1_e1v(ji,jj) - v_ice(ji,jj-1) * r1_e1v(ji,jj-1) ) * e1t(ji,jj) * e1t(ji,jj)   &
                  &   ) * r1_e1e2t(ji,jj)
               zdt2 = zdt * zdt
               
               ! --- anisotropic stress calculation --- !
               CALL update_stress_rdg (jter, nn_nevp, phi, zdiv, zdt, &
                                       zdsT, aniso_11(ji,jj), aniso_12(ji,jj), &
                                       stressptmp, stressmtmp, &
                                       stress12tmp(ji,jj), strength(ji,jj), &
                                       alphar, alphas)

               ! delta at T points
               !zdelta = SQRT( zdiv2 + ( zdt2 + zds2 ) * z1_ecc2 )  

               ! P/delta at T points
               !zp_delt(ji,jj) = strength(ji,jj) / ( zdelta + rn_creepl )

               ! alpha & beta for aEVP
               !   gamma = 0.5*P/(delta+creepl) * (c*pi)**2/Area * dt/m
               !   alpha = beta = sqrt(4*gamma)
               !IF( ln_aEVP ) THEN
               !   zalph1   = MAX( 50._wp, rpi * SQRT( 0.5_wp * zp_delt(ji,jj) * r1_e1e2t(ji,jj) * zdt_m(ji,jj) ) )
               !   z1_alph1 = 1._wp / ( zalph1 + 1._wp )
               !ENDIF
               
               ! stress at T points (zkt/=0 if landfast)
               !zs1(ji,jj) = ( zs1(ji,jj) * zalph1 + zp_delt(ji,jj) * ( zdiv * (1._wp + zkt) - zdelta * (1._wp - zkt) ) ) * z1_alph1
               !zs2(ji,jj) = ( zs2(ji,jj) * zalph2 + zp_delt(ji,jj) * ( zdt * z1_ecc2 * (1._wp + zkt) ) ) * z1_alph2
               zs1(ji,jj) = ( zs1(ji,jj) + stressptmp * zalph1  ) * z1_alph1
               zs2(ji,jj) = ( zs2(ji,jj) + stressmtmp * zalph1  ) * z1_alph1
             
            END DO
         END DO
         !CALL lbc_lnk( 'icedyn_rhg_eap', zp_delt, 'T', 1. )
         CALL lbc_lnk( 'icedyn_rhg_eap', stress12tmp, 'T', 1. )

         DO jj = 1, jpjm1
            DO ji = 1, jpim1
               ! stress12tmp at F points 
               zdsT = ( stress12tmp(ji,jj+1) * e1e2t(ji,jj+1) + stress12tmp(ji+1,jj+1) * e1e2t(ji+1,jj+1)  &
                  &   + stress12tmp(ji,jj  ) * e1e2t(ji,jj  ) + stress12tmp(ji+1,jj  ) * e1e2t(ji+1,jj  )  &
                  &   ) * 0.25_wp * r1_e1e2f(ji,jj)

               ! alpha & beta for aEVP
               !IF( ln_aEVP ) THEN
               !   zalph1   = MAX( 50._wp, rpi * SQRT( 0.5_wp * zp_delt(ji,jj) * r1_e1e2t(ji,jj) * zdt_m(ji,jj) ) )
               !   z1_alph1 = 1._wp / ( zalph1 + 1._wp )
               !ENDIF
               
               ! stress at F points (zkt/=0 if landfast)
               !zs12(ji,jj)= ( zs12(ji,jj) * zalph2 + zp_delf * ( zds(ji,jj) * z1_ecc2 * (1._wp + zkt) ) * 0.5_wp ) * z1_alph2
               zs12(ji,jj) = ( zs12(ji,jj) + stress12tmp(ji,jj) * zalph1  ) * z1_alph1

            END DO
         END DO

         ! --- Ice internal stresses (Appendix C of Hunke and Dukowicz, 2002) --- !
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1               
               !                   !--- U points
               zfU(ji,jj) = 0.5_wp * ( ( zs1(ji+1,jj) - zs1(ji,jj) ) * e2u(ji,jj)                                             &
                  &                  + ( zs2(ji+1,jj) * e2t(ji+1,jj) * e2t(ji+1,jj) - zs2(ji,jj) * e2t(ji,jj) * e2t(ji,jj)    &
                  &                    ) * r1_e2u(ji,jj)                                                                      &
                  &                  + ( zs12(ji,jj) * e1f(ji,jj) * e1f(ji,jj) - zs12(ji,jj-1) * e1f(ji,jj-1) * e1f(ji,jj-1)  &
                  &                    ) * 2._wp * r1_e1u(ji,jj)                                                              &
                  &                  ) * r1_e1e2u(ji,jj)
               !
               !                !--- V points
               zfV(ji,jj) = 0.5_wp * ( ( zs1(ji,jj+1) - zs1(ji,jj) ) * e1v(ji,jj)                                             &
                  &                  - ( zs2(ji,jj+1) * e1t(ji,jj+1) * e1t(ji,jj+1) - zs2(ji,jj) * e1t(ji,jj) * e1t(ji,jj)    &
                  &                    ) * r1_e1v(ji,jj)                                                                      &
                  &                  + ( zs12(ji,jj) * e2f(ji,jj) * e2f(ji,jj) - zs12(ji-1,jj) * e2f(ji-1,jj) * e2f(ji-1,jj)  &
                  &                    ) * 2._wp * r1_e2v(ji,jj)                                                              &
                  &                  ) * r1_e1e2v(ji,jj)
               !
               !                !--- ice currents at U-V point
               v_iceU(ji,jj) = 0.25_wp * ( v_ice(ji,jj) + v_ice(ji,jj-1) + v_ice(ji+1,jj) + v_ice(ji+1,jj-1) ) * umask(ji,jj,1)
               u_iceV(ji,jj) = 0.25_wp * ( u_ice(ji,jj) + u_ice(ji-1,jj) + u_ice(ji,jj+1) + u_ice(ji-1,jj+1) ) * vmask(ji,jj,1)
               !
            END DO
         END DO
         !
         ! --- Computation of ice velocity --- !
         !  Bouillon et al. 2013 (eq 47-48) => unstable unless alpha, beta vary as in Kimmritz 2016 & 2017
         !  Bouillon et al. 2009 (eq 34-35) => stable
         IF( MOD(jter,2) == 0 ) THEN ! even iterations
            !
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1
                  !                 !--- tau_io/(v_oce - v_ice)
                  zTauO = zaV(ji,jj) * zrhoco * SQRT( ( v_ice (ji,jj) - v_oce (ji,jj) ) * ( v_ice (ji,jj) - v_oce (ji,jj) )  &
                     &                              + ( u_iceV(ji,jj) - u_oceV(ji,jj) ) * ( u_iceV(ji,jj) - u_oceV(ji,jj) ) )
                  !                 !--- Ocean-to-Ice stress
                  ztauy_oi(ji,jj) = zTauO * ( v_oce(ji,jj) - v_ice(ji,jj) )
                  !
                  !                 !--- tau_bottom/v_ice
                  zvel  = 5.e-05_wp + SQRT( v_ice(ji,jj) * v_ice(ji,jj) + u_iceV(ji,jj) * u_iceV(ji,jj) )
                  zTauB = ztauy_base(ji,jj) / zvel
                  !                 !--- OceanBottom-to-Ice stress
                  ztauy_bi(ji,jj) = zTauB * v_ice(ji,jj)
                  !
                  !                 !--- Coriolis at V-points (energy conserving formulation)
                  zCorV(ji,jj)  = - 0.25_wp * r1_e2v(ji,jj) *  &
                     &    ( zmf(ji,jj  ) * ( e2u(ji,jj  ) * u_ice(ji,jj  ) + e2u(ji-1,jj  ) * u_ice(ji-1,jj  ) )  &
                     &    + zmf(ji,jj+1) * ( e2u(ji,jj+1) * u_ice(ji,jj+1) + e2u(ji-1,jj+1) * u_ice(ji-1,jj+1) ) )
                  !
                  !                 !--- Sum of external forces (explicit solution) = F + tau_ia + Coriolis + spg + tau_io
                  zRHS = zfV(ji,jj) + ztauy_ai(ji,jj) + zCorV(ji,jj) + zspgV(ji,jj) + ztauy_oi(ji,jj)
                  !
                  !                 !--- landfast switch => 0 = static  friction : TauB > RHS & sign(TauB) /= sign(RHS)
                  !                                         1 = sliding friction : TauB < RHS
                  rswitch = 1._wp - MIN( 1._wp, ABS( SIGN( 1._wp, zRHS + ztauy_base(ji,jj) ) - SIGN( 1._wp, zRHS ) ) )
                  !
                  IF( ln_aEVP ) THEN !--- ice velocity using aEVP (Kimmritz et al 2016 & 2017)
                     v_ice(ji,jj) = ( (          rswitch * ( zmV_t(ji,jj) * ( zbeta(ji,jj) * v_ice(ji,jj) + v_ice_b(ji,jj) )       & ! previous velocity
                        &                                  + zRHS + zTauO * v_ice(ji,jj) )                                         & ! F + tau_ia + Coriolis + spg + tau_io(only ocean part)
                        &                                  / MAX( zepsi, zmV_t(ji,jj) * ( zbeta(ji,jj) + 1._wp ) + zTauO - zTauB ) & ! m/dt + tau_io(only ice part) + landfast
                        &               + ( 1._wp - rswitch ) * v_ice(ji,jj) * MAX( 0._wp, 1._wp - zdtevp * rn_lfrelax )           & ! static friction => slow decrease to v=0
                        &             ) * zmsk01y(ji,jj) + v_oce(ji,jj) * 0.01_wp * ( 1._wp - zmsk01y(ji,jj) )                     & ! v_ice = v_oce/100 if mass < zmmin & conc < zamin
                        &           )   * zmsk00y(ji,jj)
                  ELSE               !--- ice velocity using EVP implicit formulation (cf Madec doc & Bouillon 2009)
                     v_ice(ji,jj) = ( (           rswitch   * ( zmV_t(ji,jj)  * v_ice(ji,jj)                                       & ! previous velocity
                        &                                     + zRHS + zTauO * v_ice(ji,jj) )                                      & ! F + tau_ia + Coriolis + spg + tau_io(only ocean part)
                        &                                     / MAX( zepsi, zmV_t(ji,jj) + zTauO - zTauB )                         & ! m/dt + tau_io(only ice part) + landfast
                        &                + ( 1._wp - rswitch ) * v_ice(ji,jj) * MAX( 0._wp, 1._wp - zdtevp * rn_lfrelax )          & ! static friction => slow decrease to v=0
                        &              ) * zmsk01y(ji,jj) + v_oce(ji,jj) * 0.01_wp * ( 1._wp - zmsk01y(ji,jj) )                    & ! v_ice = v_oce/100 if mass < zmmin & conc < zamin
                        &            )   * zmsk00y(ji,jj)
                  ENDIF
               END DO
            END DO
            CALL lbc_lnk( 'icedyn_rhg_eap', v_ice, 'V', -1. )
            !
#if defined key_agrif
!!            CALL agrif_interp_ice( 'V', jter, nn_nevp )
            CALL agrif_interp_ice( 'V' )
#endif
            IF( ln_bdy )   CALL bdy_ice_dyn( 'V' )
            !
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1          
                  !                 !--- tau_io/(u_oce - u_ice)
                  zTauO = zaU(ji,jj) * zrhoco * SQRT( ( u_ice (ji,jj) - u_oce (ji,jj) ) * ( u_ice (ji,jj) - u_oce (ji,jj) )  &
                     &                              + ( v_iceU(ji,jj) - v_oceU(ji,jj) ) * ( v_iceU(ji,jj) - v_oceU(ji,jj) ) )
                  !                 !--- Ocean-to-Ice stress
                  ztaux_oi(ji,jj) = zTauO * ( u_oce(ji,jj) - u_ice(ji,jj) )
                  !
                  !                 !--- tau_bottom/u_ice
                  zvel  = 5.e-05_wp + SQRT( v_iceU(ji,jj) * v_iceU(ji,jj) + u_ice(ji,jj) * u_ice(ji,jj) )
                  zTauB = ztaux_base(ji,jj) / zvel
                  !                 !--- OceanBottom-to-Ice stress
                  ztaux_bi(ji,jj) = zTauB * u_ice(ji,jj)
                  !
                  !                 !--- Coriolis at U-points (energy conserving formulation)
                  zCorU(ji,jj)  =   0.25_wp * r1_e1u(ji,jj) *  &
                     &    ( zmf(ji  ,jj) * ( e1v(ji  ,jj) * v_ice(ji  ,jj) + e1v(ji  ,jj-1) * v_ice(ji  ,jj-1) )  &
                     &    + zmf(ji+1,jj) * ( e1v(ji+1,jj) * v_ice(ji+1,jj) + e1v(ji+1,jj-1) * v_ice(ji+1,jj-1) ) )
                  !
                  !                 !--- Sum of external forces (explicit solution) = F + tau_ia + Coriolis + spg + tau_io
                  zRHS = zfU(ji,jj) + ztaux_ai(ji,jj) + zCorU(ji,jj) + zspgU(ji,jj) + ztaux_oi(ji,jj)
                  !
                  !                 !--- landfast switch => 0 = static  friction : TauB > RHS & sign(TauB) /= sign(RHS)
                  !                                         1 = sliding friction : TauB < RHS
                  rswitch = 1._wp - MIN( 1._wp, ABS( SIGN( 1._wp, zRHS + ztaux_base(ji,jj) ) - SIGN( 1._wp, zRHS ) ) )
                  !
                  IF( ln_aEVP ) THEN !--- ice velocity using aEVP (Kimmritz et al 2016 & 2017)
                     u_ice(ji,jj) = ( (          rswitch * ( zmU_t(ji,jj) * ( zbeta(ji,jj) * u_ice(ji,jj) + u_ice_b(ji,jj) )       & ! previous velocity
                        &                                  + zRHS + zTauO * u_ice(ji,jj) )                                         & ! F + tau_ia + Coriolis + spg + tau_io(only ocean part)
                        &                                  / MAX( zepsi, zmU_t(ji,jj) * ( zbeta(ji,jj) + 1._wp ) + zTauO - zTauB ) & ! m/dt + tau_io(only ice part) + landfast
                        &               + ( 1._wp - rswitch ) * u_ice(ji,jj) * MAX( 0._wp, 1._wp - zdtevp * rn_lfrelax )           & ! static friction => slow decrease to v=0
                        &             ) * zmsk01x(ji,jj) + u_oce(ji,jj) * 0.01_wp * ( 1._wp - zmsk01x(ji,jj) )                     & ! v_ice = v_oce/100 if mass < zmmin & conc < zamin 
                        &           )   * zmsk00x(ji,jj)
                  ELSE               !--- ice velocity using EVP implicit formulation (cf Madec doc & Bouillon 2009)
                     u_ice(ji,jj) = ( (           rswitch   * ( zmU_t(ji,jj)  * u_ice(ji,jj)                                       & ! previous velocity
                        &                                     + zRHS + zTauO * u_ice(ji,jj) )                                      & ! F + tau_ia + Coriolis + spg + tau_io(only ocean part)
                        &                                     / MAX( zepsi, zmU_t(ji,jj) + zTauO - zTauB )                         & ! m/dt + tau_io(only ice part) + landfast
                        &                + ( 1._wp - rswitch ) * u_ice(ji,jj) * MAX( 0._wp, 1._wp - zdtevp * rn_lfrelax )          & ! static friction => slow decrease to v=0
                        &              ) * zmsk01x(ji,jj) + u_oce(ji,jj) * 0.01_wp * ( 1._wp - zmsk01x(ji,jj) )                    & ! v_ice = v_oce/100 if mass < zmmin & conc < zamin 
                        &            )   * zmsk00x(ji,jj)
                  ENDIF
               END DO
            END DO
            CALL lbc_lnk( 'icedyn_rhg_eap', u_ice, 'U', -1. )
            !
#if defined key_agrif
!!            CALL agrif_interp_ice( 'U', jter, nn_nevp )
            CALL agrif_interp_ice( 'U' )
#endif
            IF( ln_bdy )   CALL bdy_ice_dyn( 'U' )
            !
         ELSE ! odd iterations
            !
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1
                  !                 !--- tau_io/(u_oce - u_ice)
                  zTauO = zaU(ji,jj) * zrhoco * SQRT( ( u_ice (ji,jj) - u_oce (ji,jj) ) * ( u_ice (ji,jj) - u_oce (ji,jj) )  &
                     &                              + ( v_iceU(ji,jj) - v_oceU(ji,jj) ) * ( v_iceU(ji,jj) - v_oceU(ji,jj) ) )
                  !                 !--- Ocean-to-Ice stress
                  ztaux_oi(ji,jj) = zTauO * ( u_oce(ji,jj) - u_ice(ji,jj) )
                  !
                  !                 !--- tau_bottom/u_ice
                  zvel  = 5.e-05_wp + SQRT( v_iceU(ji,jj) * v_iceU(ji,jj) + u_ice(ji,jj) * u_ice(ji,jj) )
                  zTauB = ztaux_base(ji,jj) / zvel
                  !                 !--- OceanBottom-to-Ice stress
                  ztaux_bi(ji,jj) = zTauB * u_ice(ji,jj)
                  !
                  !                 !--- Coriolis at U-points (energy conserving formulation)
                  zCorU(ji,jj)  =   0.25_wp * r1_e1u(ji,jj) *  &
                     &    ( zmf(ji  ,jj) * ( e1v(ji  ,jj) * v_ice(ji  ,jj) + e1v(ji  ,jj-1) * v_ice(ji  ,jj-1) )  &
                     &    + zmf(ji+1,jj) * ( e1v(ji+1,jj) * v_ice(ji+1,jj) + e1v(ji+1,jj-1) * v_ice(ji+1,jj-1) ) )
                  !
                  !                 !--- Sum of external forces (explicit solution) = F + tau_ia + Coriolis + spg + tau_io
                  zRHS = zfU(ji,jj) + ztaux_ai(ji,jj) + zCorU(ji,jj) + zspgU(ji,jj) + ztaux_oi(ji,jj)
                  !
                  !                 !--- landfast switch => 0 = static  friction : TauB > RHS & sign(TauB) /= sign(RHS)
                  !                                         1 = sliding friction : TauB < RHS
                  rswitch = 1._wp - MIN( 1._wp, ABS( SIGN( 1._wp, zRHS + ztaux_base(ji,jj) ) - SIGN( 1._wp, zRHS ) ) )
                  !
                  IF( ln_aEVP ) THEN !--- ice velocity using aEVP (Kimmritz et al 2016 & 2017)
                     u_ice(ji,jj) = ( (          rswitch * ( zmU_t(ji,jj) * ( zbeta(ji,jj) * u_ice(ji,jj) + u_ice_b(ji,jj) )       & ! previous velocity
                        &                                  + zRHS + zTauO * u_ice(ji,jj) )                                         & ! F + tau_ia + Coriolis + spg + tau_io(only ocean part)
                        &                                  / MAX( zepsi, zmU_t(ji,jj) * ( zbeta(ji,jj) + 1._wp ) + zTauO - zTauB ) & ! m/dt + tau_io(only ice part) + landfast
                        &               + ( 1._wp - rswitch ) * u_ice(ji,jj) * MAX( 0._wp, 1._wp - zdtevp * rn_lfrelax )           & ! static friction => slow decrease to v=0
                        &             ) * zmsk01x(ji,jj) + u_oce(ji,jj) * 0.01_wp * ( 1._wp - zmsk01x(ji,jj) )                     & ! v_ice = v_oce/100 if mass < zmmin & conc < zamin 
                        &           )   * zmsk00x(ji,jj)
                  ELSE               !--- ice velocity using EVP implicit formulation (cf Madec doc & Bouillon 2009)
                     u_ice(ji,jj) = ( (           rswitch   * ( zmU_t(ji,jj)  * u_ice(ji,jj)                                       & ! previous velocity
                        &                                     + zRHS + zTauO * u_ice(ji,jj) )                                      & ! F + tau_ia + Coriolis + spg + tau_io(only ocean part)
                        &                                     / MAX( zepsi, zmU_t(ji,jj) + zTauO - zTauB )                         & ! m/dt + tau_io(only ice part) + landfast
                        &                + ( 1._wp - rswitch ) * u_ice(ji,jj) * MAX( 0._wp, 1._wp - zdtevp * rn_lfrelax )          & ! static friction => slow decrease to v=0
                        &              ) * zmsk01x(ji,jj) + u_oce(ji,jj) * 0.01_wp * ( 1._wp - zmsk01x(ji,jj) )                    & ! v_ice = v_oce/100 if mass < zmmin & conc < zamin
                        &            )   * zmsk00x(ji,jj)
                  ENDIF
               END DO
            END DO
            CALL lbc_lnk( 'icedyn_rhg_eap', u_ice, 'U', -1. )
            !
#if defined key_agrif
!!            CALL agrif_interp_ice( 'U', jter, nn_nevp )
            CALL agrif_interp_ice( 'U' )
#endif
            IF( ln_bdy )   CALL bdy_ice_dyn( 'U' )
            !
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1
                  !                 !--- tau_io/(v_oce - v_ice)
                  zTauO = zaV(ji,jj) * zrhoco * SQRT( ( v_ice (ji,jj) - v_oce (ji,jj) ) * ( v_ice (ji,jj) - v_oce (ji,jj) )  &
                     &                              + ( u_iceV(ji,jj) - u_oceV(ji,jj) ) * ( u_iceV(ji,jj) - u_oceV(ji,jj) ) )
                  !                 !--- Ocean-to-Ice stress
                  ztauy_oi(ji,jj) = zTauO * ( v_oce(ji,jj) - v_ice(ji,jj) )
                  !
                  !                 !--- tau_bottom/v_ice
                  zvel  = 5.e-05_wp + SQRT( v_ice(ji,jj) * v_ice(ji,jj) + u_iceV(ji,jj) * u_iceV(ji,jj) )
                  zTauB = ztauy_base(ji,jj) / zvel
                  !                 !--- OceanBottom-to-Ice stress
                  ztauy_bi(ji,jj) = zTauB * v_ice(ji,jj)
                  !
                  !                 !--- Coriolis at v-points (energy conserving formulation)
                  zCorV(ji,jj)  = - 0.25_wp * r1_e2v(ji,jj) *  &
                     &    ( zmf(ji,jj  ) * ( e2u(ji,jj  ) * u_ice(ji,jj  ) + e2u(ji-1,jj  ) * u_ice(ji-1,jj  ) )  &
                     &    + zmf(ji,jj+1) * ( e2u(ji,jj+1) * u_ice(ji,jj+1) + e2u(ji-1,jj+1) * u_ice(ji-1,jj+1) ) )
                  !
                  !                 !--- Sum of external forces (explicit solution) = F + tau_ia + Coriolis + spg + tau_io
                  zRHS = zfV(ji,jj) + ztauy_ai(ji,jj) + zCorV(ji,jj) + zspgV(ji,jj) + ztauy_oi(ji,jj)
                  !
                  !                 !--- landfast switch => 0 = static  friction : TauB > RHS & sign(TauB) /= sign(RHS)
                  !                                         1 = sliding friction : TauB < RHS
                  rswitch = 1._wp - MIN( 1._wp, ABS( SIGN( 1._wp, zRHS + ztauy_base(ji,jj) ) - SIGN( 1._wp, zRHS ) ) )
                  !
                  IF( ln_aEVP ) THEN !--- ice velocity using aEVP (Kimmritz et al 2016 & 2017)
                     v_ice(ji,jj) = ( (          rswitch * ( zmV_t(ji,jj) * ( zbeta(ji,jj) * v_ice(ji,jj) + v_ice_b(ji,jj) )       & ! previous velocity
                        &                                  + zRHS + zTauO * v_ice(ji,jj) )                                         & ! F + tau_ia + Coriolis + spg + tau_io(only ocean part)
                        &                                  / MAX( zepsi, zmV_t(ji,jj) * ( zbeta(ji,jj) + 1._wp ) + zTauO - zTauB ) & ! m/dt + tau_io(only ice part) + landfast
                        &               + ( 1._wp - rswitch ) * v_ice(ji,jj) * MAX( 0._wp, 1._wp - zdtevp * rn_lfrelax )           & ! static friction => slow decrease to v=0
                        &             ) * zmsk01y(ji,jj) + v_oce(ji,jj) * 0.01_wp * ( 1._wp - zmsk01y(ji,jj) )                     & ! v_ice = v_oce/100 if mass < zmmin & conc < zamin
                        &           )   * zmsk00y(ji,jj)
                  ELSE               !--- ice velocity using EVP implicit formulation (cf Madec doc & Bouillon 2009)
                     v_ice(ji,jj) = ( (           rswitch   * ( zmV_t(ji,jj)  * v_ice(ji,jj)                                       & ! previous velocity
                        &                                     + zRHS + zTauO * v_ice(ji,jj) )                                      & ! F + tau_ia + Coriolis + spg + tau_io(only ocean part)
                        &                                     / MAX( zepsi, zmV_t(ji,jj) + zTauO - zTauB )                         & ! m/dt + tau_io(only ice part) + landfast
                        &                + ( 1._wp - rswitch ) * v_ice(ji,jj) * MAX( 0._wp, 1._wp - zdtevp * rn_lfrelax )          & ! static friction => slow decrease to v=0
                        &              ) * zmsk01y(ji,jj) + v_oce(ji,jj) * 0.01_wp * ( 1._wp - zmsk01y(ji,jj) )                    & ! v_ice = v_oce/100 if mass < zmmin & conc < zamin
                        &            )   * zmsk00y(ji,jj)
                  ENDIF
               END DO
            END DO
            CALL lbc_lnk( 'icedyn_rhg_eap', v_ice, 'V', -1. )
            !
#if defined key_agrif
!!            CALL agrif_interp_ice( 'V', jter, nn_nevp )
            CALL agrif_interp_ice( 'V' )
#endif
            IF( ln_bdy )   CALL bdy_ice_dyn( 'V' )
            !
         ENDIF

!!$         IF(ln_ctl) THEN   ! Convergence test
!!$            DO jj = 2 , jpjm1
!!$               zresr(:,jj) = MAX( ABS( u_ice(:,jj) - zu_ice(:,jj) ), ABS( v_ice(:,jj) - zv_ice(:,jj) ) )
!!$            END DO
!!$            zresm = MAXVAL( zresr( 1:jpi, 2:jpjm1 ) )
!!$            CALL mpp_max( 'icedyn_rhg_eap', zresm )   ! max over the global domain
!!$         ENDIF
         !
         !                                                ! ==================== !
      END DO                                              !  end loop over jter  !
      !                                                   ! ==================== !
      !
      !------------------------------------------------------------------------------!
      ! 4) Recompute delta, shear and div (inputs for mechanical redistribution) 
      !------------------------------------------------------------------------------!
      DO jj = 1, jpjm1
         DO ji = 1, jpim1

            ! shear at F points
            zds(ji,jj) = ( ( u_ice(ji,jj+1) * r1_e1u(ji,jj+1) - u_ice(ji,jj) * r1_e1u(ji,jj) ) * e1f(ji,jj) * e1f(ji,jj)   &
               &         + ( v_ice(ji+1,jj) * r1_e2v(ji+1,jj) - v_ice(ji,jj) * r1_e2v(ji,jj) ) * e2f(ji,jj) * e2f(ji,jj)   &
               &         ) * r1_e1e2f(ji,jj) * zfmask(ji,jj)

         END DO
      END DO           
      
      DO jj = 2, jpjm1
         DO ji = 2, jpim1 ! no vector loop
            
            ! tension**2 at T points
            zdt  = ( ( u_ice(ji,jj) * r1_e2u(ji,jj) - u_ice(ji-1,jj) * r1_e2u(ji-1,jj) ) * e2t(ji,jj) * e2t(ji,jj)   &
               &   - ( v_ice(ji,jj) * r1_e1v(ji,jj) - v_ice(ji,jj-1) * r1_e1v(ji,jj-1) ) * e1t(ji,jj) * e1t(ji,jj)   &
               &   ) * r1_e1e2t(ji,jj)
            zdt2 = zdt * zdt
            
            ! shear**2 at T points (doc eq. A16)
            zds2 = ( zds(ji,jj  ) * zds(ji,jj  ) * e1e2f(ji,jj  ) + zds(ji-1,jj  ) * zds(ji-1,jj  ) * e1e2f(ji-1,jj  )  &
               &   + zds(ji,jj-1) * zds(ji,jj-1) * e1e2f(ji,jj-1) + zds(ji-1,jj-1) * zds(ji-1,jj-1) * e1e2f(ji-1,jj-1)  &
               &   ) * 0.25_wp * r1_e1e2t(ji,jj)
            
            ! shear at T points
            pshear_i(ji,jj) = SQRT( zdt2 + zds2 )

            ! divergence at T points
            pdivu_i(ji,jj) = ( e2u(ji,jj) * u_ice(ji,jj) - e2u(ji-1,jj) * u_ice(ji-1,jj)   &
               &             + e1v(ji,jj) * v_ice(ji,jj) - e1v(ji,jj-1) * v_ice(ji,jj-1)   &
               &             ) * r1_e1e2t(ji,jj)
            
            ! delta at T points
            zdelta         = SQRT( pdivu_i(ji,jj) * pdivu_i(ji,jj) + ( zdt2 + zds2 ) * z1_ecc2 )  
            rswitch        = 1._wp - MAX( 0._wp, SIGN( 1._wp, -zdelta ) ) ! 0 if delta=0
            pdelta_i(ji,jj) = zdelta + rn_creepl * rswitch

         END DO
      END DO
      CALL lbc_lnk_multi( 'icedyn_rhg_eap', pshear_i, 'T', 1., pdivu_i, 'T', 1., pdelta_i, 'T', 1. )
      
      ! --- Store the stress tensor for the next time step --- !
      CALL lbc_lnk_multi( 'icedyn_rhg_eap', zs1, 'T', 1., zs2, 'T', 1., zs12, 'F', 1. )
      pstress1_i (:,:) = zs1 (:,:)
      pstress2_i (:,:) = zs2 (:,:)
      pstress12_i(:,:) = zs12(:,:)
      !

      !------------------------------------------------------------------------------!
      ! 5) diagnostics
      !------------------------------------------------------------------------------!
      DO jj = 1, jpj
         DO ji = 1, jpi
            zmsk00(ji,jj) = MAX( 0._wp , SIGN( 1._wp , at_i(ji,jj) - epsi06 ) ) ! 1 if ice, 0 if no ice
         END DO
      END DO

      ! --- ice-ocean, ice-atm. & ice-oceanbottom(landfast) stresses --- !
      IF(  iom_use('utau_oi') .OR. iom_use('vtau_oi') .OR. iom_use('utau_ai') .OR. iom_use('vtau_ai') .OR. &
         & iom_use('utau_bi') .OR. iom_use('vtau_bi') ) THEN
         !
         CALL lbc_lnk_multi( 'icedyn_rhg_eap', ztaux_oi, 'U', -1., ztauy_oi, 'V', -1., ztaux_ai, 'U', -1., ztauy_ai, 'V', -1., &
            &                                  ztaux_bi, 'U', -1., ztauy_bi, 'V', -1. )
         !
         CALL iom_put( 'utau_oi' , ztaux_oi * zmsk00 )
         CALL iom_put( 'vtau_oi' , ztauy_oi * zmsk00 )
         CALL iom_put( 'utau_ai' , ztaux_ai * zmsk00 )
         CALL iom_put( 'vtau_ai' , ztauy_ai * zmsk00 )
         CALL iom_put( 'utau_bi' , ztaux_bi * zmsk00 )
         CALL iom_put( 'vtau_bi' , ztauy_bi * zmsk00 )
      ENDIF
       
      ! --- divergence, shear and strength --- !
      IF( iom_use('icediv') )   CALL iom_put( 'icediv' , pdivu_i  * zmsk00 )   ! divergence
      IF( iom_use('iceshe') )   CALL iom_put( 'iceshe' , pshear_i * zmsk00 )   ! shear
      IF( iom_use('icestr') )   CALL iom_put( 'icestr' , strength * zmsk00 )   ! strength

      ! --- stress tensor --- !
      IF( iom_use('isig1') .OR. iom_use('isig2') .OR. iom_use('isig3') .OR. iom_use('normstr') .OR. iom_use('sheastr') ) THEN
         !
         ALLOCATE( zsig1(jpi,jpj) , zsig2(jpi,jpj) , zsig3(jpi,jpj) )
         !         
         DO jj = 2, jpjm1
            DO ji = 2, jpim1
               zdum1 = ( zmsk00(ji-1,jj) * pstress12_i(ji-1,jj) + zmsk00(ji  ,jj-1) * pstress12_i(ji  ,jj-1) +  &  ! stress12_i at T-point
                  &      zmsk00(ji  ,jj) * pstress12_i(ji  ,jj) + zmsk00(ji-1,jj-1) * pstress12_i(ji-1,jj-1) )  &
                  &    / MAX( 1._wp, zmsk00(ji-1,jj) + zmsk00(ji,jj-1) + zmsk00(ji,jj) + zmsk00(ji-1,jj-1) )

               zshear = SQRT( pstress2_i(ji,jj) * pstress2_i(ji,jj) + 4._wp * zdum1 * zdum1 ) ! shear stress  

               zdum2 = zmsk00(ji,jj) / MAX( 1._wp, strength(ji,jj) )

!!               zsig1(ji,jj) = 0.5_wp * zdum2 * ( pstress1_i(ji,jj) + zshear ) ! principal stress (y-direction, see Hunke & Dukowicz 2002)
!!               zsig2(ji,jj) = 0.5_wp * zdum2 * ( pstress1_i(ji,jj) - zshear ) ! principal stress (x-direction, see Hunke & Dukowicz 2002)
!!               zsig3(ji,jj) = zdum2**2 * ( ( pstress1_i(ji,jj) + strength(ji,jj) )**2 + ( rn_ecc * zshear )**2 ) ! quadratic relation linking compressive stress to shear stress
!!                                                                                                               ! (scheme converges if this value is ~1, see Bouillon et al 2009 (eq. 11))
               zsig1(ji,jj) = 0.5_wp * zdum2 * ( pstress1_i(ji,jj) )          ! compressive stress, see Bouillon et al. 2015
               zsig2(ji,jj) = 0.5_wp * zdum2 * ( zshear )                     ! shear stress
               zsig3(ji,jj) = zdum2**2 * ( ( pstress1_i(ji,jj) + strength(ji,jj) )**2 + ( rn_ecc * zshear )**2 )
            END DO
         END DO
         CALL lbc_lnk_multi( 'icedyn_rhg_eap', zsig1, 'T', 1., zsig2, 'T', 1., zsig3, 'T', 1. )
         !
         CALL iom_put( 'isig1' , zsig1 )
         CALL iom_put( 'isig2' , zsig2 )
         CALL iom_put( 'isig3' , zsig3 )
         !
         ! Stress tensor invariants (normal and shear stress N/m)
         IF( iom_use('normstr') )   CALL iom_put( 'normstr' ,       ( zs1(:,:) + zs2(:,:) )                       * zmsk00(:,:) ) ! Normal stress
         IF( iom_use('sheastr') )   CALL iom_put( 'sheastr' , SQRT( ( zs1(:,:) - zs2(:,:) )**2 + 4*zs12(:,:)**2 ) * zmsk00(:,:) ) ! Shear stress

         DEALLOCATE( zsig1 , zsig2 , zsig3 )
      ENDIF

      ! --- anisotropy tensor --- !
      IF( iom_use('aniso') ) THEN
         CALL iom_put( 'aniso' , aniso_11 * zmsk00 )
      ENDIF
      
      ! --- SIMIP --- !
      IF(  iom_use('dssh_dx') .OR. iom_use('dssh_dy') .OR. &
         & iom_use('corstrx') .OR. iom_use('corstry') .OR. iom_use('intstrx') .OR. iom_use('intstry') ) THEN
         !
         CALL lbc_lnk_multi( 'icedyn_rhg_eap', zspgU, 'U', -1., zspgV, 'V', -1., &
            &                                  zCorU, 'U', -1., zCorV, 'V', -1., zfU, 'U', -1., zfV, 'V', -1. )

         CALL iom_put( 'dssh_dx' , zspgU * zmsk00 )   ! Sea-surface tilt term in force balance (x)
         CALL iom_put( 'dssh_dy' , zspgV * zmsk00 )   ! Sea-surface tilt term in force balance (y)
         CALL iom_put( 'corstrx' , zCorU * zmsk00 )   ! Coriolis force term in force balance (x)
         CALL iom_put( 'corstry' , zCorV * zmsk00 )   ! Coriolis force term in force balance (y)
         CALL iom_put( 'intstrx' , zfU   * zmsk00 )   ! Internal force term in force balance (x)
         CALL iom_put( 'intstry' , zfV   * zmsk00 )   ! Internal force term in force balance (y)
      ENDIF

      IF(  iom_use('xmtrpice') .OR. iom_use('ymtrpice') .OR. &
         & iom_use('xmtrpsnw') .OR. iom_use('ymtrpsnw') .OR. iom_use('xatrp') .OR. iom_use('yatrp') ) THEN
         !
         ALLOCATE( zdiag_xmtrp_ice(jpi,jpj) , zdiag_ymtrp_ice(jpi,jpj) , &
            &      zdiag_xmtrp_snw(jpi,jpj) , zdiag_ymtrp_snw(jpi,jpj) , zdiag_xatrp(jpi,jpj) , zdiag_yatrp(jpi,jpj) )
         !
         DO jj = 2, jpjm1
            DO ji = 2, jpim1
               ! 2D ice mass, snow mass, area transport arrays (X, Y)
               zfac_x = 0.5 * u_ice(ji,jj) * e2u(ji,jj) * zmsk00(ji,jj)
               zfac_y = 0.5 * v_ice(ji,jj) * e1v(ji,jj) * zmsk00(ji,jj)

               zdiag_xmtrp_ice(ji,jj) = rhoi * zfac_x * ( vt_i(ji+1,jj) + vt_i(ji,jj) ) ! ice mass transport, X-component
               zdiag_ymtrp_ice(ji,jj) = rhoi * zfac_y * ( vt_i(ji,jj+1) + vt_i(ji,jj) ) !        ''           Y-   ''

               zdiag_xmtrp_snw(ji,jj) = rhos * zfac_x * ( vt_s(ji+1,jj) + vt_s(ji,jj) ) ! snow mass transport, X-component
               zdiag_ymtrp_snw(ji,jj) = rhos * zfac_y * ( vt_s(ji,jj+1) + vt_s(ji,jj) ) !          ''          Y-   ''

               zdiag_xatrp(ji,jj)     = zfac_x * ( at_i(ji+1,jj) + at_i(ji,jj) )        ! area transport,      X-component
               zdiag_yatrp(ji,jj)     = zfac_y * ( at_i(ji,jj+1) + at_i(ji,jj) )        !        ''            Y-   ''

            END DO
         END DO

         CALL lbc_lnk_multi( 'icedyn_rhg_eap', zdiag_xmtrp_ice, 'U', -1., zdiag_ymtrp_ice, 'V', -1., &
            &                                  zdiag_xmtrp_snw, 'U', -1., zdiag_ymtrp_snw, 'V', -1., &
            &                                  zdiag_xatrp    , 'U', -1., zdiag_yatrp    , 'V', -1. )

         CALL iom_put( 'xmtrpice' , zdiag_xmtrp_ice )   ! X-component of sea-ice mass transport (kg/s)
         CALL iom_put( 'ymtrpice' , zdiag_ymtrp_ice )   ! Y-component of sea-ice mass transport 
         CALL iom_put( 'xmtrpsnw' , zdiag_xmtrp_snw )   ! X-component of snow mass transport (kg/s)
         CALL iom_put( 'ymtrpsnw' , zdiag_ymtrp_snw )   ! Y-component of snow mass transport
         CALL iom_put( 'xatrp'    , zdiag_xatrp     )   ! X-component of ice area transport
         CALL iom_put( 'yatrp'    , zdiag_yatrp     )   ! Y-component of ice area transport

         DEALLOCATE( zdiag_xmtrp_ice , zdiag_ymtrp_ice , &
            &        zdiag_xmtrp_snw , zdiag_ymtrp_snw , zdiag_xatrp , zdiag_yatrp )

      ENDIF
      !
   END SUBROUTINE ice_dyn_rhg_eap

   SUBROUTINE update_stress_rdg (ksub, ndte, phi, divu, tension, &
                                 shear, a11, a12, &
                                 stressp,  stressm, &
                                 stress12, strength, &
                                 alphar, alphas)
      !!---------------------------------------------------------------------
      !!                   ***  SUBROUTINE rhg_eap_rst  ***
      !! Updates the stress depending on values of strain rate and structure
      !! tensor and for ksub=ndte it computes closing and sliding rate
      !!---------------------------------------------------------------------
      INTEGER,  INTENT(in   ) ::   ksub, ndte
      REAL(wp), INTENT(in   ) ::   phi, strength
      REAL(wp), INTENT(in   ) ::   divu, tension, shear
      REAL(wp), INTENT(in   ) ::   a11, a12
      REAL(wp), INTENT(  out) ::   stressp, stressm, stress12
      REAL(wp), INTENT(  out) ::   alphar, alphas

      INTEGER  ::   kx ,ky, ka

      REAL(wp) ::   stemp11r, stemp12r, stemp22r   
      REAL(wp) ::   stemp11s, stemp12s, stemp22s 
      REAL(wp) ::   a22, Qd11Qd11, Qd11Qd12, Qd12Qd12 
      REAL(wp) ::   Q11Q11, Q11Q12, Q12Q12 
      REAL(wp) ::   dtemp11, dtemp12, dtemp22 
      REAL(wp) ::   rotstemp11r, rotstemp12r, rotstemp22r   
      REAL(wp) ::   rotstemp11s, rotstemp12s, rotstemp22s
      REAL(wp) ::   sig11, sig12, sig22 
      REAL(wp) ::   sgprm11, sgprm12, sgprm22 
      REAL(wp) ::   invstressconviso 
      REAL(wp) ::   Angle_denom_gamma,  Angle_denom_alpha 
      REAL(wp) ::   Tany_1, Tany_2 
      REAL(wp) ::   x, y, dx, dy, da, kxw, kyw, kaw
      REAL(wp) ::   invdx, invdy, invda   
      REAL(wp) ::   dtemp1, dtemp2, atempprime, a, invsin

      REAL(wp), PARAMETER ::   kfriction = 0.45_wp
      !!---------------------------------------------------------------------
      ! 1) structure tensor
      a22 = 1._wp-a11
      Q11Q11 = 1._wp
      Q12Q12 = rsmall
      Q11Q12 = rsmall
 
      ! gamma: angle between general coordiantes and principal axis of A 
      ! here Tan2gamma = 2 a12 / (a11 - a22) 
      if((ABS(a11 - a22) > rsmall).or.(ABS(a12) > rsmall)) then      
         Angle_denom_gamma = 1._wp/sqrt( ( a11 - a22 )*( a11 - a22) + &
                             4._wp*a12*a12 )
   
         Q11Q11 = 0.5_wp + ( a11 - a22 )*0.5_wp*Angle_denom_gamma   !Cos^2 
         Q12Q12 = 0.5_wp - ( a11 - a22 )*0.5_wp*Angle_denom_gamma   !Sin^2
         Q11Q12 = a12*Angle_denom_gamma                     !CosSin 
      endif
 
      ! rotation Q*atemp*Q^T
      atempprime = Q11Q11*a11 + 2.0_wp*Q11Q12*a12 + Q12Q12*a22 
      
      ! make first principal value the largest
      atempprime = max(atempprime, 1.0_wp - atempprime)
 
      ! 2) strain rate
      dtemp11 = 0.5_wp*(divu + tension)
      dtemp12 = shear*0.5_wp
      dtemp22 = 0.5_wp*(divu - tension)

      ! here Tan2alpha = 2 dtemp12 / (dtemp11 - dtemp22) 

      Qd11Qd11 = 1.0_wp
      Qd12Qd12 = rsmall
      Qd11Qd12 = rsmall

      if((ABS( dtemp11 - dtemp22) > rsmall).or. (ABS(dtemp12) > rsmall)) then

         Angle_denom_alpha = 1.0_wp/sqrt( ( dtemp11 - dtemp22 )* &
                             ( dtemp11 - dtemp22 ) + 4.0_wp*dtemp12*dtemp12)

         Qd11Qd11 = 0.5_wp + ( dtemp11 - dtemp22 )*0.5_wp*Angle_denom_alpha !Cos^2 
         Qd12Qd12 = 0.5_wp - ( dtemp11 - dtemp22 )*0.5_wp*Angle_denom_alpha !Sin^2
         Qd11Qd12 = dtemp12*Angle_denom_alpha !CosSin
      endif

      dtemp1 = Qd11Qd11*dtemp11 + 2.0_wp*Qd11Qd12*dtemp12 + Qd12Qd12*dtemp22
      dtemp2 = Qd12Qd12*dtemp11 - 2.0_wp*Qd11Qd12*dtemp12 + Qd11Qd11*dtemp22
      ! In cos and sin values
      x = 0._wp
      if ((ABS(dtemp1) > rsmall).or.(ABS(dtemp2) > rsmall)) then
         x = atan2(dtemp2,dtemp1)
      endif      
               
      ! to ensure the angle lies between pi/4 and 9 pi/4 
      if (x < rpi*0.25_wp) x = x + rpi*2.0_wp
      
      ! y: angle between major principal axis of strain rate and structure
      ! tensor
      ! y = gamma - alpha 
      ! Expressesed componently with
      ! Tany = (Singamma*Cosgamma - Sinalpha*Cosgamma)/(Cos^2gamma - Sin^alpha)
      
      Tany_1 = Q11Q12 - Qd11Qd12
      Tany_2 = Q11Q11 - Qd12Qd12
      
      y = 0._wp
      
      if ((ABS(Tany_1) > rsmall).or.(ABS(Tany_2) > rsmall)) then
         y = atan2(Tany_1,Tany_2)
      endif
      
      ! to make sure y is between 0 and pi
      if (y > rpi) y = y - rpi
      if (y < 0)  y = y + rpi

      ! 3) update anisotropic stress tensor 
      dx   = rpi/real(nx_yield-1,kind=wp)
      dy   = rpi/real(ny_yield-1,kind=wp)
      da   = 0.5_wp/real(na_yield-1,kind=wp)
      invdx = 1._wp/dx
      invdy = 1._wp/dy
      invda = 1._wp/da

      ! % need 8 coords and 8 weights
      ! % range in kx
      kx  = int((x-rpi*0.25_wp-rpi)*invdx) + 1
      kxw = kx - (x-rpi*0.25_wp-rpi)*invdx 
      
      ky  = int(y*invdy) + 1
      kyw = ky - y*invdy 
      
      ka  = int((atempprime-0.5_wp)*invda) + 1
      kaw = ka - (atempprime-0.5_wp)*invda
      
      ! % Determine sigma_r(A1,Zeta,y) and sigma_s (see Section A1 of Tsamados 2013)
      stemp11r =  kxw   * kyw         * kaw         * s11r(kx  ,ky  ,ka  ) &
        & + (1._wp-kxw) * kyw         * kaw         * s11r(kx+1,ky  ,ka  ) &
        & + kxw         * (1._wp-kyw) * kaw         * s11r(kx  ,ky+1,ka  ) &
        & + kxw         * kyw         * (1._wp-kaw) * s11r(kx  ,ky  ,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * kaw         * s11r(kx+1,ky+1,ka  ) &
        & + (1._wp-kxw) * kyw         * (1._wp-kaw) * s11r(kx+1,ky  ,ka+1) &
        & + kxw         * (1._wp-kyw) * (1._wp-kaw) * s11r(kx  ,ky+1,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * (1._wp-kaw) * s11r(kx+1,ky+1,ka+1)
      stemp12r =  kxw   * kyw         * kaw         * s12r(kx  ,ky  ,ka  ) &
        & + (1._wp-kxw) * kyw         * kaw         * s12r(kx+1,ky  ,ka  ) &
        & + kxw         * (1._wp-kyw) * kaw         * s12r(kx  ,ky+1,ka  ) &
        & + kxw         * kyw         * (1._wp-kaw) * s12r(kx  ,ky  ,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * kaw         * s12r(kx+1,ky+1,ka  ) &
        & + (1._wp-kxw) * kyw         * (1._wp-kaw) * s12r(kx+1,ky  ,ka+1) &
        & + kxw         * (1._wp-kyw) * (1._wp-kaw) * s12r(kx  ,ky+1,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * (1._wp-kaw) * s12r(kx+1,ky+1,ka+1)
      stemp22r = kxw    * kyw         * kaw         * s22r(kx  ,ky  ,ka  ) &
        & + (1._wp-kxw) * kyw         * kaw         * s22r(kx+1,ky  ,ka  ) &
        & + kxw         * (1._wp-kyw) * kaw         * s22r(kx  ,ky+1,ka  ) &
        & + kxw         * kyw         * (1._wp-kaw) * s22r(kx  ,ky  ,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * kaw         * s22r(kx+1,ky+1,ka  ) &
        & + (1._wp-kxw) * kyw         * (1._wp-kaw) * s22r(kx+1,ky  ,ka+1) &
        & + kxw         * (1._wp-kyw) * (1._wp-kaw) * s22r(kx  ,ky+1,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * (1._wp-kaw) * s22r(kx+1,ky+1,ka+1)
      
      stemp11s =  kxw   * kyw         * kaw         * s11s(kx  ,ky  ,ka  ) &
        & + (1._wp-kxw) * kyw         * kaw         * s11s(kx+1,ky  ,ka  ) &
        & + kxw         * (1._wp-kyw) * kaw         * s11s(kx  ,ky+1,ka  ) &
        & + kxw         * kyw         * (1._wp-kaw) * s11s(kx  ,ky  ,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * kaw         * s11s(kx+1,ky+1,ka  ) &
        & + (1._wp-kxw) * kyw         * (1._wp-kaw) * s11s(kx+1,ky  ,ka+1) &
        & + kxw         * (1._wp-kyw) * (1._wp-kaw) * s11s(kx  ,ky+1,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * (1._wp-kaw) * s11s(kx+1,ky+1,ka+1)
      stemp12s =  kxw   * kyw         * kaw         * s12s(kx  ,ky  ,ka  ) &
        & + (1._wp-kxw) * kyw         * kaw         * s12s(kx+1,ky  ,ka  ) &
        & + kxw         * (1._wp-kyw) * kaw         * s12s(kx  ,ky+1,ka  ) &
        & + kxw         * kyw         * (1._wp-kaw) * s12s(kx  ,ky  ,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * kaw         * s12s(kx+1,ky+1,ka  ) &
        & + (1._wp-kxw) * kyw         * (1._wp-kaw) * s12s(kx+1,ky  ,ka+1) &
        & + kxw         * (1._wp-kyw) * (1._wp-kaw) * s12s(kx  ,ky+1,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * (1._wp-kaw) * s12s(kx+1,ky+1,ka+1)
      stemp22s =  kxw   * kyw         * kaw         * s22s(kx  ,ky  ,ka  ) &
        & + (1._wp-kxw) * kyw         * kaw         * s22s(kx+1,ky  ,ka  ) &
        & + kxw         * (1._wp-kyw) * kaw         * s22s(kx  ,ky+1,ka  ) &
        & + kxw         * kyw         * (1._wp-kaw) * s22s(kx  ,ky  ,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * kaw         * s22s(kx+1,ky+1,ka  ) &
        & + (1._wp-kxw) * kyw         * (1._wp-kaw) * s22s(kx+1,ky  ,ka+1) &
        & + kxw         * (1._wp-kyw) * (1._wp-kaw) * s22s(kx  ,ky+1,ka+1) &
        & + (1._wp-kxw) * (1._wp-kyw) * (1._wp-kaw) * s22s(kx+1,ky+1,ka+1)
                   
      ! Calculate mean ice stress over a collection of floes (Equation 3 in
      ! Tsamados 2013)

      sig11  = strength*(stemp11r + kfriction*stemp11s) * invsin
      sig12  = strength*(stemp12r + kfriction*stemp12s) * invsin
      sig22  = strength*(stemp22r + kfriction*stemp22s) * invsin

      ! Back - rotation of the stress from principal axes into general coordinates

      ! Update stress
      sgprm11 = Q11Q11*sig11 + Q12Q12*sig22 -        2._wp*Q11Q12 *sig12
      sgprm12 = Q11Q12*sig11 - Q11Q12*sig22 + (Q11Q11 - Q12Q12)*sig12
      sgprm22 = Q12Q12*sig11 + Q11Q11*sig22 +        2._wp*Q11Q12 *sig12

      stressp  = sgprm11 + sgprm22
      stress12 = sgprm12
      stressm  = sgprm11 - sgprm22

      ! Compute ridging and sliding functions in general coordinates 
      ! (Equation 11 in Tsamados 2013)
      if (ksub == ndte) then
         rotstemp11r = Q11Q11*stemp11r - 2._wp*Q11Q12* stemp12r &
                     + Q12Q12*stemp22r
         rotstemp12r = Q11Q11*stemp12r +    Q11Q12*(stemp11r-stemp22r) &
                     - Q12Q12*stemp12r
         rotstemp22r = Q12Q12*stemp11r + 2._wp*Q11Q12* stemp12r & 
                     + Q11Q11*stemp22r

         rotstemp11s = Q11Q11*stemp11s - 2._wp*Q11Q12* stemp12s &
                     + Q12Q12*stemp22s
         rotstemp12s = Q11Q11*stemp12s +    Q11Q12*(stemp11s-stemp22s) &
                     - Q12Q12*stemp12s
         rotstemp22s = Q12Q12*stemp11s + 2._wp*Q11Q12* stemp12s & 
                     + Q11Q11*stemp22s

         alphar =  rotstemp11r*dtemp11 + 2._wp*rotstemp12r*dtemp12 &
                 + rotstemp22r*dtemp22
         alphas =  rotstemp11s*dtemp11 + 2._wp*rotstemp12s*dtemp12 &
                 + rotstemp22s*dtemp22
      endif
   END SUBROUTINE update_stress_rdg 

   SUBROUTINE rhg_eap_rst( cdrw, kt )
      !!---------------------------------------------------------------------
      !!                   ***  ROUTINE rhg_eap_rst  ***
      !!                     
      !! ** Purpose :   Read or write RHG file in restart file
      !!      
      !! ** Method  :   use of IOM library
      !!----------------------------------------------------------------------
      CHARACTER(len=*) , INTENT(in) ::   cdrw   ! "READ"/"WRITE" flag
      INTEGER, OPTIONAL, INTENT(in) ::   kt     ! ice time-step
      !
      INTEGER  ::   iter                      ! local integer
      INTEGER  ::   id1, id2, id3, id4, id5   ! local integers
      INTEGER  ::   ix, iy, ip, iz, n, ia     ! local integers
    
      INTEGER, PARAMETER            ::    nz = 100
      
      REAL(wp) ::   ainit, xinit, yinit, pinit, zinit
      REAL(wp) ::   da, dx, dy, dp, dz, a1

      REAL(wp), PARAMETER           ::   eps6 = 1.0e-6_wp
      !!----------------------------------------------------------------------
      !
      IF( TRIM(cdrw) == 'READ' ) THEN        ! Read/initialize
         !                                   ! ---------------
         IF( ln_rstart ) THEN                   !* Read the restart file
            !
            id1 = iom_varid( numrir, 'stress1_i' , ldstop = .FALSE. )
            id2 = iom_varid( numrir, 'stress2_i' , ldstop = .FALSE. )
            id3 = iom_varid( numrir, 'stress12_i', ldstop = .FALSE. )
            id4 = iom_varid( numrir, 'aniso_11' , ldstop = .FALSE. )
            id5 = iom_varid( numrir, 'aniso_12' , ldstop = .FALSE. )
            !
            IF( MIN( id1, id2, id3, id4, id5 ) > 0 ) THEN      ! fields exist
               CALL iom_get( numrir, jpdom_autoglo, 'stress1_i' , stress1_i  )
               CALL iom_get( numrir, jpdom_autoglo, 'stress2_i' , stress2_i  )
               CALL iom_get( numrir, jpdom_autoglo, 'stress12_i', stress12_i )
               CALL iom_get( numrir, jpdom_autoglo, 'aniso_11' , aniso_11 )
               CALL iom_get( numrir, jpdom_autoglo, 'aniso_12' , aniso_12 )
            ELSE                                     ! start rheology from rest
               IF(lwp) WRITE(numout,*)
               IF(lwp) WRITE(numout,*) '   ==>>>   previous run without rheology, set stresses to 0'
               stress1_i (:,:) = 0._wp
               stress2_i (:,:) = 0._wp
               stress12_i(:,:) = 0._wp
               aniso_11 (:,:) = 0.5_wp
               aniso_12 (:,:) = 0._wp
            ENDIF
         ELSE                                   !* Start from rest
            IF(lwp) WRITE(numout,*)
            IF(lwp) WRITE(numout,*) '   ==>>>   start from rest: set stresses to 0'
            stress1_i (:,:) = 0._wp
            stress2_i (:,:) = 0._wp
            stress12_i(:,:) = 0._wp
            aniso_11 (:,:) = 0.5_wp
            aniso_12 (:,:) = 0._wp
         ENDIF
         !

      da = 0.5_wp/real(na_yield-1,kind=wp)
      ainit = 0.5_wp - da
      dx = rpi/real(nx_yield-1,kind=wp)
      xinit = rpi + 0.25_wp*rpi - dx
      dz = rpi/real(nz,kind=wp)
      zinit = -rpi*0.5_wp
      dy = rpi/real(ny_yield-1,kind=wp)
      yinit = -dy

      DO ia=1,na_yield
       DO ix=1,nx_yield
        DO iy=1,ny_yield
         s11r(ix,iy,ia) = 0._wp
         s12r(ix,iy,ia) = 0._wp
         s22r(ix,iy,ia) = 0._wp
         s11s(ix,iy,ia) = 0._wp
         s12s(ix,iy,ia) = 0._wp
         s22s(ix,iy,ia) = 0._wp
         IF (ia <= na_yield-1) then
          DO iz=1,nz
          s11r(ix,iy,ia) = s11r(ix,iy,ia) + 1*w1(ainit+ia*da)* &
           exp(-w2(ainit+ia*da)*(zinit+iz*dz)*(zinit+iz*dz))* &
           s11kr(xinit+ix*dx,yinit+iy*dy,zinit+iz*dz,phi)*dz/sin(2._wp*phi)
          s12r(ix,iy,ia) = s12r(ix,iy,ia) + 1*w1(ainit+ia*da)* &
           exp(-w2(ainit+ia*da)*(zinit+iz*dz)*(zinit+iz*dz))* &
           s12kr(xinit+ix*dx,yinit+iy*dy,zinit+iz*dz,phi)*dz/sin(2._wp*phi)
          s22r(ix,iy,ia) = s22r(ix,iy,ia) + 1*w1(ainit+ia*da)* &
           exp(-w2(ainit+ia*da)*(zinit+iz*dz)*(zinit+iz*dz))* &
           s22kr(xinit+ix*dx,yinit+iy*dy,zinit+iz*dz,phi)*dz/sin(2._wp*phi) 
          s11s(ix,iy,ia) = s11s(ix,iy,ia) + 1*w1(ainit+ia*da)* &
           exp(-w2(ainit+ia*da)*(zinit+iz*dz)*(zinit+iz*dz))* &
           s11ks(xinit+ix*dx,yinit+iy*dy,zinit+iz*dz,phi)*dz/sin(2._wp*phi)
          s12s(ix,iy,ia) = s12s(ix,iy,ia) + 1*w1(ainit+ia*da)* &
           exp(-w2(ainit+ia*da)*(zinit+iz*dz)*(zinit+iz*dz))* &
           s12ks(xinit+ix*dx,yinit+iy*dy,zinit+iz*dz,phi)*dz/sin(2._wp*phi)
          s22s(ix,iy,ia) = s22s(ix,iy,ia) + 1*w1(ainit+ia*da)* &
           exp(-w2(ainit+ia*da)*(zinit+iz*dz)*(zinit+iz*dz))* &
           s22ks(xinit+ix*dx,yinit+iy*dy,zinit+iz*dz,phi)*dz/sin(2._wp*phi)
          ENDDO
          IF (abs(s11r(ix,iy,ia)) < eps6) s11r(ix,iy,ia) = 0._wp
          IF (abs(s12r(ix,iy,ia)) < eps6) s12r(ix,iy,ia) = 0._wp
          IF (abs(s22r(ix,iy,ia)) < eps6) s22r(ix,iy,ia) = 0._wp
          IF (abs(s11s(ix,iy,ia)) < eps6) s11s(ix,iy,ia) = 0._wp
          IF (abs(s12s(ix,iy,ia)) < eps6) s12s(ix,iy,ia) = 0._wp
          IF (abs(s22s(ix,iy,ia)) < eps6) s22s(ix,iy,ia) = 0._wp
         ELSE
          s11r(ix,iy,ia) = 0.5_wp*s11kr(xinit+ix*dx,yinit+iy*dy,0._wp,phi)/sin(2._wp*phi)
          s12r(ix,iy,ia) = 0.5_wp*s12kr(xinit+ix*dx,yinit+iy*dy,0._wp,phi)/sin(2._wp*phi)
          s22r(ix,iy,ia) = 0.5_wp*s22kr(xinit+ix*dx,yinit+iy*dy,0._wp,phi)/sin(2._wp*phi)
          s11s(ix,iy,ia) = 0.5_wp*s11ks(xinit+ix*dx,yinit+iy*dy,0._wp,phi)/sin(2._wp*phi)
          s12s(ix,iy,ia) = 0.5_wp*s12ks(xinit+ix*dx,yinit+iy*dy,0._wp,phi)/sin(2._wp*phi)
          s22s(ix,iy,ia) = 0.5_wp*s22ks(xinit+ix*dx,yinit+iy*dy,0._wp,phi)/sin(2._wp*phi)
          IF (abs(s11r(ix,iy,ia)) < eps6) s11r(ix,iy,ia) = 0._wp
          IF (abs(s12r(ix,iy,ia)) < eps6) s12r(ix,iy,ia) = 0._wp
          IF (abs(s22r(ix,iy,ia)) < eps6) s22r(ix,iy,ia) = 0._wp
          IF (abs(s11s(ix,iy,ia)) < eps6) s11s(ix,iy,ia) = 0._wp
          IF (abs(s12s(ix,iy,ia)) < eps6) s12s(ix,iy,ia) = 0._wp
          IF (abs(s22s(ix,iy,ia)) < eps6) s22s(ix,iy,ia) = 0._wp
         ENDIF
        ENDDO
       ENDDO
      ENDDO

      ELSEIF( TRIM(cdrw) == 'WRITE' ) THEN   ! Create restart file
         !                                   ! -------------------
         IF(lwp) WRITE(numout,*) '---- rhg-rst ----'
         iter = kt + nn_fsbc - 1             ! ice restarts are written at kt == nitrst - nn_fsbc + 1
         !
         CALL iom_rstput( iter, nitrst, numriw, 'stress1_i' , stress1_i  )
         CALL iom_rstput( iter, nitrst, numriw, 'stress2_i' , stress2_i  )
         CALL iom_rstput( iter, nitrst, numriw, 'stress12_i', stress12_i )
         CALL iom_rstput( iter, nitrst, numriw, 'aniso_11' , aniso_11 )
         CALL iom_rstput( iter, nitrst, numriw, 'aniso_12' , aniso_12 )
         !
      ENDIF
      !
   END SUBROUTINE rhg_eap_rst


   FUNCTION w1(a)
      !!-------------------------------------------------------------------
      !! Function : w1 (see Gaussian function psi in Tsamados et al 2013)
      !!-------------------------------------------------------------------
      REAL(wp), INTENT(in   ) ::   a
      REAL(wp) ::   w1
      !!-------------------------------------------------------------------
   
      w1 = - 223.87569446_wp &
           + 2361.2198663_wp*a &
           - 10606.56079975_wp*a*a &
           + 26315.50025642_wp*a*a*a &
           - 38948.30444297_wp*a*a*a*a &
           + 34397.72407466_wp*a*a*a*a*a &
           - 16789.98003081_wp*a*a*a*a*a*a &
           + 3495.82839237_wp*a*a*a*a*a*a*a
   
   END FUNCTION w1


   FUNCTION w2(a)
      !!-------------------------------------------------------------------
      !! Function : w2 (see Gaussian function psi in Tsamados et al 2013)
      !!-------------------------------------------------------------------
      REAL(wp), INTENT(in   ) ::   a
      REAL(wp) ::   w2
      !!-------------------------------------------------------------------
   
      w2 = - 6670.68911883_wp &
           + 70222.33061536_wp*a &
           - 314871.71525448_wp*a*a &
           + 779570.02793492_wp*a*a*a &
           - 1151098.82436864_wp*a*a*a*a &
           + 1013896.59464498_wp*a*a*a*a*a &
           - 493379.44906738_wp*a*a*a*a*a*a &
           + 102356.551518_wp*a*a*a*a*a*a*a
   
   END FUNCTION w2

   FUNCTION s11kr(x,y,z,phi) 
      !!-------------------------------------------------------------------
      !! Function : s11kr
      !!-------------------------------------------------------------------
      REAL(wp), INTENT(in   ) ::   x,y,z,phi
   
      REAL(wp) ::   s11kr, p, pih
   
      REAL(wp) ::   n1t2i11, n1t2i12, n1t2i21, n1t2i22
      REAL(wp) ::   n2t1i11, n2t1i12, n2t1i21, n2t1i22
      REAL(wp) ::   t1t2i11, t1t2i12, t1t2i21, t1t2i22
      REAL(wp) ::   t2t1i11, t2t1i12, t2t1i21, t2t1i22
      REAL(wp) ::   d11, d12, d22
      REAL(wp) ::   IIn1t2, IIn2t1, IIt1t2
      REAL(wp) ::   Hen1t2, Hen2t1
      !!-------------------------------------------------------------------
   
      pih = 0.5_wp*rpi
      p = phi
   
      n1t2i11 = cos(z+pih-p) * cos(z+p)
      n1t2i12 = cos(z+pih-p) * sin(z+p)
      n1t2i21 = sin(z+pih-p) * cos(z+p)
      n1t2i22 = sin(z+pih-p) * sin(z+p)
      n2t1i11 = cos(z-pih+p) * cos(z-p)
      n2t1i12 = cos(z-pih+p) * sin(z-p)
      n2t1i21 = sin(z-pih+p) * cos(z-p)
      n2t1i22 = sin(z-pih+p) * sin(z-p)
      t1t2i11 = cos(z-p) * cos(z+p)
      t1t2i12 = cos(z-p) * sin(z+p)
      t1t2i21 = sin(z-p) * cos(z+p)
      t1t2i22 = sin(z-p) * sin(z+p)
      t2t1i11 = cos(z+p) * cos(z-p)
      t2t1i12 = cos(z+p) * sin(z-p)
      t2t1i21 = sin(z+p) * cos(z-p)
      t2t1i22 = sin(z+p) * sin(z-p)
   ! In expression of tensor d, with this formulatin d(x)=-d(x+pi)
   ! Solution, when diagonalizing always check sgn(a11-a22) if > then keep x else
   ! x=x-pi/2
      d11 = cos(y)*cos(y)*(cos(x)+sin(x)*tan(y)*tan(y))
      d12 = cos(y)*cos(y)*tan(y)*(-cos(x)+sin(x))
      d22 = cos(y)*cos(y)*(sin(x)+cos(x)*tan(y)*tan(y))
      IIn1t2 = n1t2i11 * d11 + (n1t2i12 + n1t2i21) * d12 + n1t2i22 * d22
      IIn2t1 = n2t1i11 * d11 + (n2t1i12 + n2t1i21) * d12 + n2t1i22 * d22
      IIt1t2 = t1t2i11 * d11 + (t1t2i12 + t1t2i21) * d12 + t1t2i22 * d22
   
      IF (-IIn1t2>=rsmall) then
      Hen1t2 = 1._wp
      ELSE
      Hen1t2 = 0._wp
      ENDIF
   
      IF (-IIn2t1>=rsmall) then
      Hen2t1 = 1._wp
      ELSE
      Hen2t1 = 0._wp
      ENDIF
   
      s11kr = (- Hen1t2 * n1t2i11 - Hen2t1 * n2t1i11)

   END FUNCTION s11kr

   FUNCTION s12kr(x,y,z,phi)
      !!-------------------------------------------------------------------
      !! Function : s12kr
      !!-------------------------------------------------------------------
      REAL(wp), INTENT(in   ) ::   x,y,z,phi

      REAL(wp) ::   s12kr, s12r0, s21r0, p, pih

      REAL(wp) ::   n1t2i11, n1t2i12, n1t2i21, n1t2i22
      REAL(wp) ::   n2t1i11, n2t1i12, n2t1i21, n2t1i22
      REAL(wp) ::   t1t2i11, t1t2i12, t1t2i21, t1t2i22
      REAL(wp) ::   t2t1i11, t2t1i12, t2t1i21, t2t1i22
      REAL(wp) ::   d11, d12, d22
      REAL(wp) ::   IIn1t2, IIn2t1, IIt1t2
      REAL(wp) ::   Hen1t2, Hen2t1
      !!-------------------------------------------------------------------
      pih = 0.5_wp*rpi
      p = phi
   
      n1t2i11 = cos(z+pih-p) * cos(z+p)
      n1t2i12 = cos(z+pih-p) * sin(z+p)
      n1t2i21 = sin(z+pih-p) * cos(z+p)
      n1t2i22 = sin(z+pih-p) * sin(z+p)
      n2t1i11 = cos(z-pih+p) * cos(z-p)
      n2t1i12 = cos(z-pih+p) * sin(z-p)
      n2t1i21 = sin(z-pih+p) * cos(z-p)
      n2t1i22 = sin(z-pih+p) * sin(z-p)
      t1t2i11 = cos(z-p) * cos(z+p)
      t1t2i12 = cos(z-p) * sin(z+p)
      t1t2i21 = sin(z-p) * cos(z+p)
      t1t2i22 = sin(z-p) * sin(z+p)
      t2t1i11 = cos(z+p) * cos(z-p)
      t2t1i12 = cos(z+p) * sin(z-p)
      t2t1i21 = sin(z+p) * cos(z-p)
      t2t1i22 = sin(z+p) * sin(z-p)
      d11 = cos(y)*cos(y)*(cos(x)+sin(x)*tan(y)*tan(y))
      d12 = cos(y)*cos(y)*tan(y)*(-cos(x)+sin(x))
      d22 = cos(y)*cos(y)*(sin(x)+cos(x)*tan(y)*tan(y))
      IIn1t2 = n1t2i11 * d11 + (n1t2i12 + n1t2i21) * d12 + n1t2i22 * d22
      IIn2t1 = n2t1i11 * d11 + (n2t1i12 + n2t1i21) * d12 + n2t1i22 * d22
      IIt1t2 = t1t2i11 * d11 + (t1t2i12 + t1t2i21) * d12 + t1t2i22 * d22
   
      IF (-IIn1t2>=rsmall) then
      Hen1t2 = 1._wp
      ELSE
      Hen1t2 = 0._wp
      ENDIF
   
      IF (-IIn2t1>=rsmall) then
      Hen2t1 = 1._wp
      ELSE
      Hen2t1 = 0._wp
      ENDIF
   
      s12r0 = (- Hen1t2 * n1t2i12 - Hen2t1 * n2t1i12)
      s21r0 = (- Hen1t2 * n1t2i21 - Hen2t1 * n2t1i21)
      s12kr=0.5_wp*(s12r0+s21r0)
   
   END FUNCTION s12kr

   FUNCTION s22kr(x,y,z,phi)
      !!-------------------------------------------------------------------
      !! Function : s22kr
      !!-------------------------------------------------------------------
      REAL(wp), INTENT(in   ) ::   x,y,z,phi

      REAL(wp) ::   s22kr, p, pih

      REAL(wp) ::   n1t2i11, n1t2i12, n1t2i21, n1t2i22
      REAL(wp) ::   n2t1i11, n2t1i12, n2t1i21, n2t1i22
      REAL(wp) ::   t1t2i11, t1t2i12, t1t2i21, t1t2i22
      REAL(wp) ::   t2t1i11, t2t1i12, t2t1i21, t2t1i22
      REAL(wp) ::   d11, d12, d22
      REAL(wp) ::   IIn1t2, IIn2t1, IIt1t2
      REAL(wp) ::   Hen1t2, Hen2t1
      !!-------------------------------------------------------------------
      pih = 0.5_wp*rpi
      p = phi

      n1t2i11 = cos(z+pih-p) * cos(z+p)
      n1t2i12 = cos(z+pih-p) * sin(z+p)
      n1t2i21 = sin(z+pih-p) * cos(z+p)
      n1t2i22 = sin(z+pih-p) * sin(z+p)
      n2t1i11 = cos(z-pih+p) * cos(z-p)
      n2t1i12 = cos(z-pih+p) * sin(z-p)
      n2t1i21 = sin(z-pih+p) * cos(z-p)
      n2t1i22 = sin(z-pih+p) * sin(z-p)
      t1t2i11 = cos(z-p) * cos(z+p)
      t1t2i12 = cos(z-p) * sin(z+p)
      t1t2i21 = sin(z-p) * cos(z+p)
      t1t2i22 = sin(z-p) * sin(z+p)
      t2t1i11 = cos(z+p) * cos(z-p)
      t2t1i12 = cos(z+p) * sin(z-p)
      t2t1i21 = sin(z+p) * cos(z-p)
      t2t1i22 = sin(z+p) * sin(z-p)
      d11 = cos(y)*cos(y)*(cos(x)+sin(x)*tan(y)*tan(y))
      d12 = cos(y)*cos(y)*tan(y)*(-cos(x)+sin(x))
      d22 = cos(y)*cos(y)*(sin(x)+cos(x)*tan(y)*tan(y))
      IIn1t2 = n1t2i11 * d11 + (n1t2i12 + n1t2i21) * d12 + n1t2i22 * d22
      IIn2t1 = n2t1i11 * d11 + (n2t1i12 + n2t1i21) * d12 + n2t1i22 * d22
      IIt1t2 = t1t2i11 * d11 + (t1t2i12 + t1t2i21) * d12 + t1t2i22 * d22

      IF (-IIn1t2>=rsmall) then
      Hen1t2 = 1._wp
      ELSE
      Hen1t2 = 0._wp
      ENDIF

      IF (-IIn2t1>=rsmall) then
      Hen2t1 = 1._wp
      ELSE
      Hen2t1 = 0._wp
      ENDIF

      s22kr = (- Hen1t2 * n1t2i22 - Hen2t1 * n2t1i22)

   END FUNCTION s22kr

   FUNCTION s11ks(x,y,z,phi)
      !!-------------------------------------------------------------------
      !! Function : s11ks
      !!-------------------------------------------------------------------
      REAL(wp), INTENT(in   ) ::   x,y,z,phi

      REAL(wp) ::   s11ks, p, pih

      REAL(wp) ::   n1t2i11, n1t2i12, n1t2i21, n1t2i22
      REAL(wp) ::   n2t1i11, n2t1i12, n2t1i21, n2t1i22
      REAL(wp) ::   t1t2i11, t1t2i12, t1t2i21, t1t2i22
      REAL(wp) ::   t2t1i11, t2t1i12, t2t1i21, t2t1i22
      REAL(wp) ::   d11, d12, d22
      REAL(wp) ::   IIn1t2, IIn2t1, IIt1t2
      REAL(wp) ::   Hen1t2, Hen2t1
      !!-------------------------------------------------------------------
      pih = 0.5_wp*rpi
      p = phi

      n1t2i11 = cos(z+pih-p) * cos(z+p)
      n1t2i12 = cos(z+pih-p) * sin(z+p)
      n1t2i21 = sin(z+pih-p) * cos(z+p)
      n1t2i22 = sin(z+pih-p) * sin(z+p)
      n2t1i11 = cos(z-pih+p) * cos(z-p)
      n2t1i12 = cos(z-pih+p) * sin(z-p)
      n2t1i21 = sin(z-pih+p) * cos(z-p)
      n2t1i22 = sin(z-pih+p) * sin(z-p)
      t1t2i11 = cos(z-p) * cos(z+p)
      t1t2i12 = cos(z-p) * sin(z+p)
      t1t2i21 = sin(z-p) * cos(z+p)
      t1t2i22 = sin(z-p) * sin(z+p)
      t2t1i11 = cos(z+p) * cos(z-p)
      t2t1i12 = cos(z+p) * sin(z-p)
      t2t1i21 = sin(z+p) * cos(z-p)
      t2t1i22 = sin(z+p) * sin(z-p)
      d11 = cos(y)*cos(y)*(cos(x)+sin(x)*tan(y)*tan(y))
      d12 = cos(y)*cos(y)*tan(y)*(-cos(x)+sin(x))
      d22 = cos(y)*cos(y)*(sin(x)+cos(x)*tan(y)*tan(y))
      IIn1t2 = n1t2i11 * d11 + (n1t2i12 + n1t2i21) * d12 + n1t2i22 * d22
      IIn2t1 = n2t1i11 * d11 + (n2t1i12 + n2t1i21) * d12 + n2t1i22 * d22
      IIt1t2 = t1t2i11 * d11 + (t1t2i12 + t1t2i21) * d12 + t1t2i22 * d22

      IF (-IIn1t2>=rsmall) then
      Hen1t2 = 1._wp
      ELSE
      Hen1t2 = 0._wp
      ENDIF

      IF (-IIn2t1>=rsmall) then
      Hen2t1 = 1._wp
      ELSE
      Hen2t1 = 0._wp
      ENDIF

      s11ks = sign(1._wp,IIt1t2+rsmall)*(Hen1t2 * t1t2i11 + Hen2t1 * t2t1i11)

   END FUNCTION s11ks

   FUNCTION s12ks(x,y,z,phi)
      !!-------------------------------------------------------------------
      !! Function : s12ks
      !!-------------------------------------------------------------------
      REAL(wp), INTENT(in   ) ::   x,y,z,phi

      REAL(wp) ::   s12ks, s12s0, s21s0, p, pih

      REAL(wp) ::   n1t2i11, n1t2i12, n1t2i21, n1t2i22
      REAL(wp) ::   n2t1i11, n2t1i12, n2t1i21, n2t1i22
      REAL(wp) ::   t1t2i11, t1t2i12, t1t2i21, t1t2i22
      REAL(wp) ::   t2t1i11, t2t1i12, t2t1i21, t2t1i22
      REAL(wp) ::   d11, d12, d22
      REAL(wp) ::   IIn1t2, IIn2t1, IIt1t2
      REAL(wp) ::   Hen1t2, Hen2t1
      !!-------------------------------------------------------------------
      pih = 0.5_wp*rpi
      p =phi

      n1t2i11 = cos(z+pih-p) * cos(z+p)
      n1t2i12 = cos(z+pih-p) * sin(z+p)
      n1t2i21 = sin(z+pih-p) * cos(z+p)
      n1t2i22 = sin(z+pih-p) * sin(z+p)
      n2t1i11 = cos(z-pih+p) * cos(z-p)
      n2t1i12 = cos(z-pih+p) * sin(z-p)
      n2t1i21 = sin(z-pih+p) * cos(z-p)
      n2t1i22 = sin(z-pih+p) * sin(z-p)
      t1t2i11 = cos(z-p) * cos(z+p)
      t1t2i12 = cos(z-p) * sin(z+p)
      t1t2i21 = sin(z-p) * cos(z+p)
      t1t2i22 = sin(z-p) * sin(z+p)
      t2t1i11 = cos(z+p) * cos(z-p)
      t2t1i12 = cos(z+p) * sin(z-p)
      t2t1i21 = sin(z+p) * cos(z-p)
      t2t1i22 = sin(z+p) * sin(z-p)
      d11 = cos(y)*cos(y)*(cos(x)+sin(x)*tan(y)*tan(y))
      d12 = cos(y)*cos(y)*tan(y)*(-cos(x)+sin(x))
      d22 = cos(y)*cos(y)*(sin(x)+cos(x)*tan(y)*tan(y))
      IIn1t2 = n1t2i11 * d11 + (n1t2i12 + n1t2i21) * d12 + n1t2i22 * d22
      IIn2t1 = n2t1i11 * d11 + (n2t1i12 + n2t1i21) * d12 + n2t1i22 * d22
      IIt1t2 = t1t2i11 * d11 + (t1t2i12 + t1t2i21) * d12 + t1t2i22 * d22

      IF (-IIn1t2>=rsmall) then
      Hen1t2 = 1._wp
      ELSE
      Hen1t2 = 0._wp
      ENDIF

      IF (-IIn2t1>=rsmall) then
      Hen2t1 = 1._wp
      ELSE
      Hen2t1 = 0._wp
      ENDIF

      s12s0 = sign(1._wp,IIt1t2+rsmall)*(Hen1t2 * t1t2i12 + Hen2t1 * t2t1i12)
      s21s0 = sign(1._wp,IIt1t2+rsmall)*(Hen1t2 * t1t2i21 + Hen2t1 * t2t1i21)
      s12ks=0.5_wp*(s12s0+s21s0)

   END FUNCTION s12ks

   FUNCTION s22ks(x,y,z,phi) 
      !!-------------------------------------------------------------------
      !! Function : s22ks
      !!-------------------------------------------------------------------
      REAL(wp), INTENT(in   ) ::   x,y,z,phi

      REAL(wp) ::   s22ks, p, pih

      REAL(wp) ::   n1t2i11, n1t2i12, n1t2i21, n1t2i22
      REAL(wp) ::   n2t1i11, n2t1i12, n2t1i21, n2t1i22
      REAL(wp) ::   t1t2i11, t1t2i12, t1t2i21, t1t2i22
      REAL(wp) ::   t2t1i11, t2t1i12, t2t1i21, t2t1i22
      REAL(wp) ::   d11, d12, d22
      REAL(wp) ::   IIn1t2, IIn2t1, IIt1t2
      REAL(wp) ::   Hen1t2, Hen2t1
      !!-------------------------------------------------------------------
      pih = 0.5_wp*rpi
      p =phi

      n1t2i11 = cos(z+pih-p) * cos(z+p)
      n1t2i12 = cos(z+pih-p) * sin(z+p)
      n1t2i21 = sin(z+pih-p) * cos(z+p)
      n1t2i22 = sin(z+pih-p) * sin(z+p)
      n2t1i11 = cos(z-pih+p) * cos(z-p)
      n2t1i12 = cos(z-pih+p) * sin(z-p)
      n2t1i21 = sin(z-pih+p) * cos(z-p)
      n2t1i22 = sin(z-pih+p) * sin(z-p)
      t1t2i11 = cos(z-p) * cos(z+p)
      t1t2i12 = cos(z-p) * sin(z+p)
      t1t2i21 = sin(z-p) * cos(z+p)
      t1t2i22 = sin(z-p) * sin(z+p)
      t2t1i11 = cos(z+p) * cos(z-p)
      t2t1i12 = cos(z+p) * sin(z-p)
      t2t1i21 = sin(z+p) * cos(z-p)
      t2t1i22 = sin(z+p) * sin(z-p)
      d11 = cos(y)*cos(y)*(cos(x)+sin(x)*tan(y)*tan(y))
      d12 = cos(y)*cos(y)*tan(y)*(-cos(x)+sin(x))
      d22 = cos(y)*cos(y)*(sin(x)+cos(x)*tan(y)*tan(y))
      IIn1t2 = n1t2i11 * d11 + (n1t2i12 + n1t2i21) * d12 + n1t2i22 * d22
      IIn2t1 = n2t1i11 * d11 + (n2t1i12 + n2t1i21) * d12 + n2t1i22 * d22
      IIt1t2 = t1t2i11 * d11 + (t1t2i12 + t1t2i21) * d12 + t1t2i22 * d22

      IF (-IIn1t2>=rsmall) THEN
      Hen1t2 = 1._wp
      ELSE
      Hen1t2 = 0._wp
      ENDIF

      IF (-IIn2t1>=rsmall) THEN
      Hen2t1 = 1._wp
      ELSE
      Hen2t1 = 0._wp
      ENDIF

      s22ks = sign(1._wp,IIt1t2+rsmall)*(Hen1t2 * t1t2i22 + Hen2t1 * t2t1i22)

   END FUNCTION s22ks

#else
   !!----------------------------------------------------------------------
   !!   Default option         Empty module           NO SI3 sea-ice model
   !!----------------------------------------------------------------------

#endif
   !!==============================================================================
END MODULE icedyn_rhg_eap
